/** \file listener.h
 * \brief TINE static listener documentation 
 *
 */
#ifndef LISTENER_H_
#define LISTENER_H_

#include "project.def"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#if defined(TBUFSRVDLL_EXPORTS) || defined(TBUFSRV64_EXPORTS)
# define EXPORT_B __declspec(dllexport) 
#endif 
#if defined(TBUFSRVDLL_IMPORTS) || defined(TBUFSRV64_IMPORTS)
# define EXPORT_B __declspec(dllimport) 
#endif
#ifndef EXPORT_B
# define EXPORT_B 
#endif

#define ALLOWED_IDLE_TIME 300
typedef struct mexListenerStruct
{
  char devName[132];
  char devProperty[PROPERTY_NAME_SIZE];
  DTYPE dout;
  DTYPE din;
  BYTE *outbuf;
  BYTE *inbuf;
  BYTE **outarr;   /* output arrays for volatile data */
  int capacity;    /* output capacity for volatile data */
  int mlid;
  int lid;         /* link id */
  int nid;         /* notifier id */
  void (*nf)(int); /* notifier function */
  int status;
  int idleTimeInSeconds;
  int pollingRate;
  int isGlobal;
  time_t lasttime;
  double timestamp;
  struct mexListenerStruct *nxt;
} mexListener;

#ifndef MAX_NUM_LISTENERS
#define MAX_NUM_LISTENERS 5000
#endif
extern mexListener **MexListenerTable;
extern int nMexListeners;
extern int MexListenerTableCapacity;

/* setDefaultListenerMode() */
/**
 * \brief Sets the polling mode for any static listeners created via a call to
 * alsnExecLink().
 *
 * The polling mode to be used in establishing an asynchronous static listener is
 * not a input parameter to the api call alsnExecLink().  Instead the polling mode
 * which is applied is determined by the current value of the default listener mode.
 * (CM_POLL upon startup).  In any case the CM_WAIT flag is applied the the poll mode
 * so that the synchronous call alsnExecLink() does not return until the initial call
 * has completed.
 * 
 * \param mode is the desired default polling mode, i.e. one of CM_POLL or CM_REFRESH
 * (possibly ORed with CM_CONNECT or CM_NETWORK for instance).  Setting the default 
 * mode to CM_SINGLE or CM_CANCEL makes no sense in the context of a static listener.
 *
 * \return The current default polling mode to be applied to a static listener.
 * 
 */
EXPORT_B int setDefaultListenerMode(int mode);
int stopListener(mexListener *ml);
mexListener *startListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int pollrate,int flags);
int stopListener(mexListener *ml);
void stopAllListeners(void);
mexListener *getListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int rate);
int getDataFromListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,double *dTime,int rate,int flags);
int isGlobal(char *devName);
int isMetaCall(char *devName,char *devProperty);
int isQueryProperty(char *devProperty);
EXPORT_B int getPropertyInfoFromListener(char *dev, char *prp, int *siz, int *fmt);

/* alsnExecLink() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. For query calls (for instance to obtain a property
 * list), calls to the local history server, or calls with a timeout parameter less than 50 
 * milliseconds, the call reverts to ExecLinkEx(), i.e. synchronous data transfer.  Otherwise,
 * the initial call starts an asynchronous static listener for the selected device name and
 * property, polling at the rate determined by the timeout parameter. 
 * The initial call waits for the first data set to be returned before it completes.  Subsequent
 * calls retrieve data from the listener's buffer and complete immediately.  If no subsequent
 * calls are made within a 5 minute interval, the asynchronous link is cancelled.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the output data set.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the input data set.
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly.  This parameter
 * has a particular meaning when used in the static listener routines.  Namely, it is also used to
 * determine the polling interval for the asynchronous listener which is attached to the call.  You
 * can force the call to execute synchronously without starting a listener by passing a value less
 * than 50 for this parameter.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by examining
 * RPCLastErrorString.
 *
 * \see ExecLink(), ExecLinkEx(), DTYPE
 */
EXPORT_B int alsnExecLink(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access,long timeout);
EXPORT_B int alsnExecLinkEx(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access,long timeout,double *dTimeStamp);

/* alsnGetLastLinkError() */
/**
 * \brief The error string associated with the input error number
 * 
 * \param cc is the completion code for which the error string is desired.  If cc is in fact
 * the last Link Error then the returned error string is used, otherwise the system error
 * string for the completion code in question is returned
 * \param errstr is the error string buffer which is to contain the string description of the
 * the cc parameter.  A pointer to the same string buffer is returned (A null pointer returns "[null error string]").
 *  
 * \return The error string associated with the input error number.
 *
 */
EXPORT_B char *alsnGetLastLinkError(short cc, char *errstr);

EXPORT_B short alsnCall(char *fullDeviceNameAndProperty,DTYPE *dout,DTYPE *din,short access,double *dTimeStamp);
EXPORT_B short tputgetArrayEx(char *fullDeviceNameAndProperty, short access,
                              short putdatafmt,void *putdata,int putsiz,
                              short getdatafmt, void *getdata, int getsiz,double *dTimeStamp);
EXPORT_B short tputString(char *fullDeviceNameAndProperty,char *putdata);
EXPORT_B short tputStringArray(char *fullDeviceNameAndProperty,char *putdata,int putsize);
EXPORT_B short tputDoubleArray(char *fullDeviceNameAndProperty,double *putdata,int putsiz);
EXPORT_B short tputSingleArray(char *fullDeviceNameAndProperty,float *putdata,int putsiz);
EXPORT_B short tputLongArray(char *fullDeviceNameAndProperty,int *putdata,int putsiz);
EXPORT_B short tputShortArray(char *fullDeviceNameAndProperty,short *putdata,int putsiz);
EXPORT_B short tAttachNotifier(char *fullDeviceNameAndProperty, void (*nf)(int), int nid);
EXPORT_B void tDetachNotifier(void (*nf)(int));
EXPORT_B short tgetArray(char *fullDeviceNameAndProperty,short datafmt,void *dataout,int buffersiz);
EXPORT_B short tgetSingleArray(char *fullDeviceNameAndProperty,float *dataout,int buffersiz);
EXPORT_B short tputgetSingleArray(char *fullDeviceNameAndProperty,float *datain,int insize,float *dataout,int outsize,double *datatime);
EXPORT_B short tgetDoubleArray(char *fullDeviceNameAndProperty,double *dataout,int buffersiz);
EXPORT_B short tputgetDoubleArray(char *fullDeviceNameAndProperty,double *datain,int insize,double *dataout,int outsize,double *datatime);
EXPORT_B short tgetLongArray(char *fullDeviceNameAndProperty,long *dataout,int buffersiz);
EXPORT_B short tputgetLongArray(char *fullDeviceNameAndProperty,long *datain,int insize,long *dataout,int outsize,double *datatime);
EXPORT_B short tgetShortArray(char *fullDeviceNameAndProperty,short *dataout,int buffersiz);
EXPORT_B short tputgetShortArray(char *fullDeviceNameAndProperty,short *datain,int insize,short *dataout,int outsize,double *datatime);
EXPORT_B short tgetStringArray(char *fullDeviceNameAndProperty,char *dataout,int buffersiz);
EXPORT_B short tgetString(char *fullDeviceNameAndProperty,char *dataout,int buffersiz);
EXPORT_B short tputgetString(char *fullDeviceNameAndProperty,char *datain,char *dataout,int buffersiz);
EXPORT_B short tputgetStringEx(char *fullDeviceNameAndProperty,char *datain,char *dataout,int buffersiz,double *datatime);

extern mexListener *mexListeners;
extern int nListeners;

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* LISTENER_H_ */
