/** \file srvcore.h
 *
 * server-side (and client-side) transport definitions and prototypes.
 *
 */
#ifndef SRVCORE_H_
#define SRVCORE_H_

#include "project.def"
#include "prolog.h"
#include "tinetype.h"
#include "errors.h"
#include "almlib.h"
#include "threader.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifndef MULTITHREADED
#define gRunServerCycleInSeparateThread FALSE
#define gRunTransportInSeparateThread FALSE
#define ReleaseSystemSemaphore(m) (m++)
#define WaitForSystemSemaphore(m,to) (m > 0 ? m-- : 0)
#define ReleaseSystemMutex(m) (m = 0)
#define WaitForSystemMutex(m,to) ((m++) > 0 ? -1 : 0)
extern int hSystemClientMutex;
extern int hSystemGlobalsUpdateMutex;
extern int hSystemServerMutex;
extern int hSystemKernelMutex;
extern int hLinkTblMutex;
extern int hDpdLinkMutex;
extern int hLwdLinkMutex;
extern int hLinkLstMutex;
extern int hLinkQueueMutex;
extern int hClientTableMutex;
#else
void ReleaseSyncLink(int flag);
extern int gRunServerCycleInSeparateThread;
extern int gRunTransportInSeparateThread;
extern int gNeedToStartSrvCycleThread;
extern int gNeedToStartClnCycleThread;
extern int gNeedToStartEqmThreads;
extern int gNeedToStopEqmThreads;
#endif
/* SetRunServerCycleInSeparateThread() */
/**
 * \fn TINE_EXPORT void SetRunServerCycleInSeparateThread(int value)
 * \brief Determines whether the server cycle is run in its own thread (boolean).
 * 
 * If set to TRUE (default for multi-threaded builds), the server cycle will run in its
 * own thread, separate from the client cycle.  If set to FALSE, both cycles will run
 * in a round-robin fashion, where all socket descriptor sets are 'select()ed' on to assure
 * a prompt response.  This is usually fine.  However under extreme scenarios where performance
 * is of primary concern, then leaving this setting in its default TRUE state is the best 
 * policy.  This of course only applies to multi-threaded builds of the TINE library.
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.  TRUE is
 * the default.
 *
 * \see GetRunServerCycleInSeparateThread().
 */
TINE_EXPORT void SetRunServerCycleInSeparateThread(int value);
/* GetRunServerCycleInSeparateThread() */
/**
 * \fn TINE_EXPORT int GetRunServerCycleInSeparateThread(void)
 * \brief Returns whether the server cycle is run in its own thread (boolean).
 * 
 * If set to TRUE (default for multi-threaded builds), the server cycle will run in its
 * own thread, separate from the client cycle.  If set to FALSE, both cycles will run
 * in a round-robin fashion, where all socket descriptor sets are 'select()ed' on to assure
 * a prompt response.  This is usually fine.  However under extreme scenarios where performance
 * is of primary concern, then leaving this setting in its default TRUE state is the best 
 * policy.  This of course only applies to multi-threaded builds of the TINE library.
 *
 * \return the current setting of this feature.
 *
 * \see SetRunServerCycleInSeparateThread().
 */
TINE_EXPORT int GetRunServerCycleInSeparateThread(void);
/* SetRunTransportInSeparateThread() */
/**
 * \fn TINE_EXPORT void SetRunTransportInSeparateThread(int value)
 * \brief Determines whether the (tcp) transport is run in its own thread (boolean).
 * 
 * If set to TRUE (default for most multi-threaded builds), connected sockets will each have
 * their own thread where prompt data transport is assured.  If set to FALSE, all transport
 * delivery will follow the server contract access.  This (setting to FALSE) is generally an okay solution 
 * and is not very resource intensive.  Setting this to TRUE will cause multiple buffering
 * of data payloads for each connected peer, which could cause memory problems on some OSes such as
 * VxWorks or other systems whether there is more concern about memory than speed.
 * The default setting is TRUE for windows, unix, and macos builds, and FALSE for VxWorks.
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.  FALSE is
 * the default.
 *
 * \see GetRunServerCycleInSeparateThread().
 */
TINE_EXPORT void SetRunTransportInSeparateThread(int value);
/* GetRunTransportInSeparateThread() */
/**
 * \fn TINE_EXPORT int GetRunTransportInSeparateThread(void)
 * \brief Returns whether the (tcp) transport is run in its own thread (boolean).
 * 
 * If set to TRUE (default for most multi-threaded builds), connected sockets will each have
 * their own thread where prompt data transport is assured.  If set to FALSE, all transport
 * delivery will follow the server contract access.  This (setting to FALSE) is generally an okay solution 
 * and is not very resource intensive.  Setting this to TRUE will cause multiple buffering
 * of data payloads for each connected peer, which could cause memory problems on some OSes such as
 * VxWorks or other systems whether there is more concern about memory than speed.
 * The default setting is TRUE for windows, unix, and macos builds, and FALSE for VxWorks.
 *
 * \return the current setting of this feature.
 *
 * \see SetRunTransportInSeparateThread().
 */
TINE_EXPORT int GetRunTransportInSeparateThread(void);

extern int gAllowBkgTaskReentrancy;
/* SetAllowBackgroundTaskReentrancy() */
/**
 * \fn TINE_EXPORT void SetAllowBackgroundTaskReentrancy(int value)
 * \brief Determines whether equipment module background tasks may be re-entered (boolean).
 * 
 * If set to TRUE, the server's background task will allow re-entrancy
 * which might occur if the task rate instructs the background task to be called
 * even though the previous call has not yet returned.  In such cases the programmer
 * must deal with whatever reentrancy issues there might be.  The default is value is FALSE.
 * (not reentrant).
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.  FALSE is
 * the default.
 *
 * \see GetAllowBackgroundTaskReentrancy().
 */
TINE_EXPORT void SetAllowBackgroundTaskReentrancy(int value);

/* GetAllowBackgroundTaskReentrancy() */
/**
 * \fn TINE_EXPORT int GetAllowBackgroundTaskReentrancy(void)
 * \brief Returns whether equipment module background tasks may be re-entered (boolean).
 * 
 * If set to TRUE, the server's background task will allow re-entrancy
 * which might occur if the task rate instructs the background task to be called
 * even though the previous call has not yet returned.  In such cases the programmer
 * must deal with whatever reentrancy issues there might be.  This routine returns the
 * current setting.
 *
 * \return the current setting of this feature.
 *
 * \see SetAllowBackgroundTaskReentrancy().
 */
TINE_EXPORT int GetAllowBackgroundTaskReentrancy(void);

/**
 * \fn TINE_EXPORT void SetUseMultiThreadedBackgroundTasks(int value)
 * \brief Determines whether equipment module background tasks are to run in separate threads (boolean).
 * 
 * If set to TRUE, the server's background task will run in its own thread, where a multi-threaded
 * build of the library is in use.  Note that this could result in data concurrency issues if the
 * background task is acquiring data which is to be used by an equipment module's equipment function.
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off. 
 * TRUE is the default.
 *
 * \see SetAllowBackgroundTaskReentrancy(), GetUseMultiThreadedBackgroundTasks().
 */
TINE_EXPORT void SetUseMultiThreadedBackgroundTasks(int value);

/**
 * \fn TINE_EXPORT int GetUseMultiThreadedBackgroundTasks(void)
 * \brief Returns whether equipment module background tasks are to run in separate threads (boolean).
 * 
 * If set to TRUE, the server's background task will run in its own thread, where a multi-threaded
 * build of the library is in use.  Note that this could result in data concurrency issues if the
 * background task is acquiring data which is to be used by an equipment module's equipment function.
 *
 * \return the current setting of this feature.
 *
 * \see SetAllowBackgroundTaskReentrancy(), SetUseMultiThreadedBackgroundTasks().
 */
TINE_EXPORT int GetUseMultiThreadedBackgroundTasks(void);

/**
 * \fn TINE_EXPORT void SetUseMultiThreadedEquipmentFunctions(int value)
 * \brief Determines whether an equipment module equipment function can run in a separate threads (boolean).
 * 
 * If set to TRUE, the an equipment function will handle designated properties in a separate thread, 
 * where a multi-threaded build of the library is in use.  
 * Ordinarily an equipment function should be regarded as an interrupt service routine which should return
 * 'promptly' after being called.  All properties will be directed to the equipment function using the same
 * thread as the server cycle, unless otherwise specified.  On the other hand if a property is specifically
 * registered to run the equipment module in a separate thread (because the action is known to require 
 * significant time to complete) then it will run in its own thread if this setting has been enabled.
 * This will result in other calls to the equipment function receiving an 'operation busy' completion
 * while the equipment function is handling the time-consuming call.  Otherwise, concurrent calls will 
 * simply time out at the client side.  To configure a property to be called in a separate thread one can
 * either call SetCallPropertyInSeparateThread() following the property registration of supply an "MT" switch
 * in the properties 'access' string which registering via file (either exports.csv or fec.xml).
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.  
 * FALSE is the default.
 *
 * \see GetUseMultiThreadedEquipmentFunctions(), SetUseMultiThreadedStockFunctions(), SetCallPropertyInSeparateThread().
 */
TINE_EXPORT void SetUseMultiThreadedEquipmentFunctions(int value);

/**
 * \fn TINE_EXPORT int GetUseMultiThreadedEquipmentFunctions(void)
 * \brief Returns whether an equipment module equipment function can run in a separate threads (boolean).
 * 
 * If set to TRUE, the an equipment function will handle designated properties in a separate thread, 
 * where a multi-threaded build of the library is in use.  
 * Ordinarily an equipment function should be regarded as an interrupt service routine which should return
 * 'promptly' after being called.  All properties will be directed to the equipment function using the same
 * thread as the server cycle, unless otherwise specified.  On the other hand if a property is specifically
 * registered to run the equipment module in a separate thread (because the action is known to require 
 * significant time to complete) then it will run in its own thread if this setting has been enabled.
 * This will result in other calls to the equipment function receiving an 'operation busy' completion
 * while the equipment function is handling the time-consuming call.  Otherwise, concurrent calls will 
 * simply time out at the client side.  To configure a property to be called in a separate thread one can
 * either call SetCallPropertyInSeparateThread() following the property registration of supply an "MT" switch
 * in the properties 'access' string which registering via file (either exports.csv or fec.xml).
 *
 * \return the current setting of this feature.
 *
 * \see SetUseMultiThreadedEquipmentFunctions(), SetUseMultiThreadedStockFunctions(), SetCallPropertyInSeparateThread().
 */
TINE_EXPORT int GetUseMultiThreadedEquipmentFunctions(void);

/**
 * \fn TINE_EXPORT void SetUseMultiThreadedStockFunctions(int value)
 * \brief Determines whether stock propery calls can run in a separate threads (boolean).
 * 
 * If set to TRUE, certain stock property calls will be handled in a separate thread, 
 * where a multi-threaded build of the library is in use.  
 * Ordinarily stock property calls return 'promptly' after being called, and make use of the same
 * thread as the server cycle, unless otherwise specified.  On the other hand certain stock properties 
 * which make heavy use of the local disk might require considerable time to complete, noteably calls to
 * the local history subsystem or file retrieval calls.
 * A calls to such property will run in its own thread if this setting has been enabled.
 * This will result in other calls to the same stock property receiving an 'operation busy' completion
 * while the system is handling the time-consuming call.  Otherwise, concurrent calls will 
 * simply time out at the client side.  
 *
 * \param value should be non-zero (TRUE) to turn this feature on or FALSE to turn it off.  
 * TRUE is the default.
 *
 * \see GetUseMultiThreadedStockFunctions().
 */
TINE_EXPORT void SetUseMultiThreadedStockFunctions(int value);

/**
 * \fn TINE_EXPORT int GetUseMultiThreadedStockFunctions(void)
 * \brief Returns whether stock propery calls can run in a separate threads (boolean).
 * 
 * If set to TRUE, certain stock property calls will be handled in a separate thread, 
 * where a multi-threaded build of the library is in use.  
 * Ordinarily stock property calls return 'promptly' after being called, and make use of the same
 * thread as the server cycle, unless otherwise specified.  On the other hand certain stock properties 
 * which make heavy use of the local disk might require considerable time to complete, noteably calls to
 * the local history subsystem or file retrieval calls.
 * A calls to such property will run in its own thread if this setting has been enabled.
 * This will result in other calls to the same stock property receiving an 'operation busy' completion
 * while the system is handling the time-consuming call.  Otherwise, concurrent calls will 
 * simply time out at the client side.  
 *
 * \return the current setting of this feature.
 *
 * \see SetUseMultiThreadedStockFunctions().
 */
TINE_EXPORT int GetUseMultiThreadedStockFunctions(void);

/**
 * \fn TINE_EXPORT int SetServerThreadPriority(int priority)
 * \brief Determines the priority of the server cycle thread as well as other associated server-side threads.
 * 
 * The priority passed should be regarded as an offset to the 'normal' thread priorty for the operating
 * system in use.  That is, the TINE definition STD_THREAD_PRIORITY will be 'correct' for the target operating
 * system (where a multi-threaded build of the library is in use).  It is up to the developer to take the
 * appropriate steps to ensure that this call in fact works as planned.  For instance, using real-time priorities
 * on Windows will require a call to the Windows function SetPriorityClass() using REALTIME_PRIORITY_CLASS prior
 * to server initialization (otherwise the resulting thread priorities will not be as desired).  Using any 
 * non-standard priorities on Unix will likely require running the application as root!  VxWorks will take the
 * initial task priority as a reference for the 'standard thread priority'.
 *
 * \param priority is the desired thread priority (default STD_THREAD_PRIORITY).  
 *
 * \return the assigned server thread priority.
 *
 * \see GetServerThreadPriority().
 */
TINE_EXPORT int SetServerThreadPriority(int priority);

/**
 * \fn TINE_EXPORT int GetServerThreadPriority(void)
 * \brief Returns the priority of the server cycle thread as well as other associated server-side threads.
 * 
 * \return the server thread priority.
 *
 * \see SetServerThreadPriority().
 */
TINE_EXPORT int GetServerThreadPriority(void);

/**
 * \fn TINE_EXPORT int SetClientThreadPriority(int priority)
 * \brief Determines the priority of the client cycle thread as well as other associated client-side threads.
 * 
 * The priority passed should be regarded as an offset to the 'normal' thread priorty for the operating
 * system in use.  That is, the TINE definition STD_THREAD_PRIORITY will be 'correct' for the target operating
 * system (where a multi-threaded build of the library is in use).  It is up to the developer to take the
 * appropriate steps to ensure that this call in fact works as planned.  For instance, using real-time priorities
 * on Windows will require a call to the Windows function SetPriorityClass() using REALTIME_PRIORITY_CLASS prior
 * to server initialization (otherwise the resulting thread priorities will not be as desired).  Using any 
 * non-standard priorities on Unix will likely require running the application as root!  VxWorks will take the
 * initial task priority as a reference for the 'standard thread priority'.
 *
 * \param priority is the desired thread priority (default STD_THREAD_PRIORITY).  
 *
 * \return the assigned client thread priority.
 *
 * \see GetClientThreadPriority().
 */
TINE_EXPORT int SetClientThreadPriority(int priority);

/**
 * \fn TINE_EXPORT int GetClientThreadPriority(void)
 * \brief Returns the priority of the client cycle thread as well as other associated client-side threads.
 * 
 * \return the client thread priority.
 *
 * \see SetClientThreadPriority().
 */
TINE_EXPORT int GetClientThreadPriority(void);

/**
 * \fn TINE_EXPORT int SetBackgroundThreadPriority(int priority)
 * \brief Determines the priority of any registered background threads.
 * 
 * The priority passed should be regarded as an offset to the 'normal' thread priorty for the operating
 * system in use.  That is, the TINE definition STD_THREAD_PRIORITY will be 'correct' for the target operating
 * system (where a multi-threaded build of the library is in use).  It is up to the developer to take the
 * appropriate steps to ensure that this call in fact works as planned.  For instance, using real-time priorities
 * on Windows will require a call to the Windows function SetPriorityClass() using REALTIME_PRIORITY_CLASS prior
 * to server initialization (otherwise the resulting thread priorities will not be as desired).  Using any 
 * non-standard priorities on Unix will likely require running the application as root!  VxWorks will take the
 * initial task priority as a reference for the 'standard thread priority'.
 *
 * \param priority is the desired thread priority (default STD_THREAD_PRIORITY).  
 *
 * \return the assigned background thread priority.
 *
 * \see GetBackgroundThreadPriority().
 */
TINE_EXPORT int SetBackgroundThreadPriority(int priority);

/**
 * \fn TINE_EXPORT int GetBackgroundThreadPriority(void)
 * \brief Returns the priority of the registered background threads
 * 
 * \return the background thread priority.
 *
 * \see SetBackgroundThreadPriority().
 */
TINE_EXPORT int GetBackgroundThreadPriority(void);

/**
 * \fn TINE_EXPORT int SetKernelPriority(int priority)
 * \brief Determines the priority of TINE kernel threads.
 * 
 * Use this routine to set the priority of all TINE kernel threads (server-side, client-side,
 * background tasks) to the desired value.  This is primarily a convenience routine and is
 * equivalent to calling SetServerThreadPriority(), SetClientThreadPriority() and 
 * SetBackgroundThreadPriority() individually.
 *
 * \param priority is the desired thread priority (default STD_THREAD_PRIORITY).  
 *
 * \return the assigned thread priority.
 *
 * \see SetServerThreadPriority(), SetClientThreadPriority(), SetBackgroundThreadPriority()
 */
TINE_EXPORT int SetKernelPriority(int priority);

int WaitForSyncLink(int id,int to);
int InitSystemMutexSet(void);

/* lockEquipmentModules() */
/**
 * \fn TINE_EXPORT int LockEquipmentModules(void)
 * \brief Locks all equipment modules
 * 
 * Applying this call seizes a system mutex which will not allow the scheduler to cycle
 * thru the equipment module dispatch routines.
 * Consider setting the IDLE state of the module in lieu
 * of locking and unlocking.
 *  
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see UnlockEquipmentModules().
 */
TINE_EXPORT int LockEquipmentModules(void);

/* UnlockEquipmentModules() */
/**
 * \fn TINE_EXPORT int UnlockEquipmentModules(void)
 * \brief Unlocks all equipment modules
 * 
 * If the equipment modules were locked by applying a call to LockEquipmentModules()
 * then this call frees the lock state, allowing equipment module dispatch routines and
 * background tasks to continue.  Consider setting the IDLE state of the module in lieu
 * of locking and unlocking.
 *  
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see LockEquipmentModules().
 */
TINE_EXPORT int UnlockEquipmentModules(void);

#define DEFAULT_TRANSPORT_PROTOCOL_LEVEL 6

/**
 * \internal
 * \brief Default value for MaxNumConnections.
 */
#ifndef CONTBL_CAPACITY
# define CONTBL_CAPACITY 1000 /* connection table entries  */
#endif

/**
 * \internal
 * \brief Communication protocol macro definitions ...
 */
#define IPX             0x01
#define SPX             0x02
#define TCP             0x04
#define UDP             0x08
#define PIPE            0x10
#define MMF             0x20
#define STREAM          0x40
#define DBGPIPE         0x80

#define getProtocolAsString(i) (\
  (i) == IPX ? "IPX" :\
  (i) == SPX ? "SPX" :\
  (i) == TCP ? "TCP" :\
  (i) == UDP ? "UDP" :\
  (i) == PIPE ? "LCL" :\
  (i) == MMF ? "LCL" :\
  (i) == STREAM ? "STREAM" :\
  (i) == DBGPIPE ? "DBGPIPE" :\
  "---"\
)

#define CLN_REQ_ 0x01
#define SRV_RSP_ 0x02

#define GENS_TAG "GRPEQM"
#define ENS_TAG "ENSEQM"
#define CAS_TAG "CASEQM"
#define GENS_FEC "GENS"

#define USERS_LABEL "USERS"

#define FAILOVER_MASTER_DESC "failover MASTER"
#define FAILOVER_SLAVE_DESC "failover SLAVE"

#define FAILOVER_NONE   0x00
#define FAILOVER_MASTER 0x01
#define FAILOVER_SLAVE  0x02

#define FAILOVER_THRESHOLD 5
#define FAILOVER_POLLING_RATE 1000

/**
 * \internal
 * \brief Tine Engine operation mode
 *
 * Single process OSes (DOS, NIOS) use POLLING, where in some sense, TINE becomes the OS.  
 * Multi-process OSes (UNIX, WINDOWS, VxWorks, VMS, etc.) use SELECTING, where the
 * TINE engines only wakes up if 'select()' fires (incoming data) or if the 
 * SystemPolling rate so dictates.
 */
#define POLLING         0x01
#define SELECTING       0x02

/**
 * \internal
 * \brief The default value for MaxPollingRate.
 *
 * Client contracts asking to poll properties at an interval greater than MaxPollingRate will
 * be automatically converted to MaxPollingRate.
 * This in effect sets the upper limit for the timeout parameter for select() when SELECTING
 * mode is in effect.  Note that 'Rate' is actually a misnomer as the value of this parameter
 * actually gives an interval and not a rate.
 *
 */
#ifndef DEFAULT_POLLRATE
#  ifndef SYSPOLL
#    define DEFAULT_POLLRATE 10
#  else
#    define DEFAULT_POLLRATE SYSPOLL
#  endif
#endif

/**
 * \internal
 * \brief The default value for MinPollingRate.
 *
 * Client contracts asking to poll properties at an interval smaller than MinPollingRate will
 * be automatically converted to MinPollingRate.  Thus a renegade client cannot force a server
 * to poll a property at smaller intervals.
 *
 */
#ifdef WIN32
# define MIN_POLLRATE 10
#endif
#ifndef MIN_POLLRATE
#  ifndef SYSPOLL
#    define MIN_POLLRATE 20
#  else
#    define MIN_POLLRATE SYSPOLL
#  endif
#endif

/**
 * \internal
 * \brief The default timeout used when ExecLink() is called.
 *
 * ExecLink() calls ExecLinkEx() with DEFAULT_CLIENT_POLLRATE as the timeout parameter.
 *
 */
#define DEFAULT_CLIENT_POLLRATE 1000

/**
 * \internal
 * \brief The default value for MaxSystemTransportSize
 *
 * Buffer space is reserved at initialization according to this value (64 KBytes).  
 * Servers which need to support larger data transfers should set this value
 * accordingly.
 *
 * \todo Make this gracefully dynamic if possible. (Watch out for VxWorks, NIOS).
 *
 */
#define MAX_TRANSPORTSIZE 0xffff

 /**
 * \internal
 * \brief The default value of gBurstLimit.
 *
 * This corresponds to the number of data packets sent without a break.
 * If UDP is in use, there is no flow control, so data transfers are made assuming that
 * all clients can handle the network speed of the server.  
 * This may or may not be true. The default value is a generally reasonable choice, but
 * Video servers sending large blocks of data will probably want to set gBurstLimit to a 
 * high value on initialization.
 *
 */
#ifndef BURST_LIMIT
# define BURST_LIMIT 1000
#endif
extern int gBurstLimit;

 /**
 * \internal
 * \brief The default value of gCycleDelay.
 *
 * This corresponds to the number of milliseonds to delay following a burst transmission.
 * If UDP is in use, there is no flow control, so data transfers are made assuming that
 * all clients can handle the network speed of the server.  
 * This may or may not be true. The default value is a generally reasonable choice, but
 * Video servers sending large blocks of data will probably want to set gCycleDelay to a 
 * smaller value (or 0) on initialization.
 *
 */
#ifndef CYCLE_DELAY
# define CYCLE_DELAY   20
#endif
extern int gCycleDelay;

/**
 * \internal
 * \brief The default value of gSystemTick.
 *
 * Used internally within the TINE engine scheduler.
 *
 */
#ifndef SYSTEM_TICK
# define SYSTEM_TICK 10
#endif
extern UINT32 gSystemTick;

/**
 * \internal
 * \brief The default value of gSynchronizeContracts.
 *
 * Determines whether the TINE engine scheduler should attempt to
 * force contracts with polling intervals of (modulo) 1000 msec
 * should be called consecutively within the same cycle.  This
 * greatly improves the 'packing' of contracts being returned to
 * a given client.
 *
 */
#ifndef CONTRACT_SYNCHRONIZE
# define CONTRACT_SYNCHRONIZE TRUE
#endif
extern UINT32 gSynchronizeContracts;
extern int gSystemExitCode;
extern int gSystemRunningStandAlone;
extern int gSchedulePropertyLazy;
extern int gClientBurstLimit;
extern int gAutoLinkWatchdogs;
extern int allowCommonLinks;
extern int gAlwaysRetry;
extern int gServerCycleIdle;

/**
 * \internal
 * \brief Used as a flag in lieu of the startup time field in a contract.
 *
 * A server servicing a NETWORK subscription will set the startup time 
 * field in the contract header to BCAST_ID (for either a multicast or
 * broadcast request) as a special signature.
 *
 */
#define BCAST_ID 255

/**
 * \internal
 * \brief Special scheduling signals for the TINE scheduler.
 *
 */
#define PRP_SCHEDULE_SIGNAL 0
#define PRP_REQUEST_SIGNAL 1
#define PRP_CANCEL_SIGNAL 2

/**
 * \internal
 * \brief Special reserved system names ...
 *
 */
#define SRVEQM_NAME "_SRV__"
#define SRVFEC_NAME "NETWORK"
#define SRVEXP_NAME "NETWORK"
#define STKEQM_NAME "_STK__"
/**
 * \internal
 * \brief Determine whether this request is coming from a NETWORK request
 * for services (e.g. ENS, CAS, etc.).
 *
 */
#define isServiceRequest(pctr) (strcmp((pctr)->EqmName,SRVEQM_NAME) == 0 ? TRUE : FALSE)

/**
 * \internal
 * \brief The default value for gRespondToServiceRequests.
 *
 * Determine whether this server can respond to Service requests.
 * Typically only the ENS will have this features turned on.  (Or possibly the CAS) 
 *
 */
#ifndef RESPOND_TO_SERVICE_REQUESTS
# define RESPOND_TO_SERVICE_REQUESTS FALSE
#endif
extern TINE_EXPORT int gRespondToServiceRequests;

#define REGISTERED_ENS    0x01
#define REGISTERED_GENS   0x02
#define REGISTERED_MASTER 0x04
#define REGISTERED_ALL    0x07

#define TYPE_PRP 0x01
#define TYPE_DEV 0x02

TINE_EXPORT void SetForwardWildcardCalls(int type,int value);
int fwdWildCardCall(int type,int fmt);

typedef struct EQPTask
{
  void (*tskfcn)(void); /* background task function */
  int weight;           /* background task weight factor */
  struct EQPTask *next;
} EQPTASK;
extern EQPTASK *EQPTaskList;

#define EQPBKG void (*)(void)
#define EQPINI void (*)(void)
#define EQPEXI void (*)(void)

#define ALLOCSIZE(b) ((b) + (8 - (b) % 8))

/**
 * \internal
 * \brief The top level header which defines the incoming data packet.
 * 
 * \todo split totalSize away from the header.  This will involve a good look
 * at the routines in swaplib.c.  The advantage of doing this is primarily to 
 * deal with a PktHdr with fields on a 4-byte boundary.  The first two
 * bytes of what has come in doesn't need to be a part of this header. BUT,
 * splitting this off from the header has low priority.
 *
 */
typedef struct /* struct is passed over the net */
{
  UINT16 totalSize; /**< total packet size in bytes (must match the bytes read) */
  char userName[USERNAME_SIZE]; /**< caller name */
  short tineProtocol; /**< tine protocol level (modern: 6 for contracts, 4 for globals) */
  UINT16 number; /**< tine version number (contracts) or number of incoming keywords (globals) */
} PktHdr;
#define PKTHDR_SIZE  22
#define SUB_PROT_LOCATION 18

#define LNKTBL_HASH_SIZE 211

#define REVISIONID(m,r) ((m)*256 + (r))

/**
 * \internal
 * \brief callback prototypes
 */
typedef void (*TCBFCNP)(int,int);
typedef void (*XTCBFCNP)(int,int,void *);

#define TCBCALL(ce,cc) {\
  ce->isInsideCallback = TRUE;\
  ce->hasReference ?\
  ((XTCBFCNP)ce->cbfcn)((int)ce->cbId,cc,ce->cbRef) :\
  ((TCBFCNP)ce->cbfcn)((int)ce->cbId,cc);\
  ce->isInsideCallback = FALSE;\
}

typedef struct BlackLnkLstStruct
{
  char key[FULL_NAME_MAXSIZE+8];
  int  status;
  struct BlackLnkLstStruct *nxt;
} BLACKLISTED_LINK;
extern BLACKLISTED_LINK *gBlackLnkTbl[LNKTBL_HASH_SIZE];
extern int gBlackLnkTblSize;
BLACKLISTED_LINK *getBlackLnkLstItem(char *ctx, char *srv, char *dev, char *prp);
void appendBlackLnkLst(char *ctx, char *srv, char *dev, char *prp,int status);
void flushBlackLnkLst(void);

#define RELINKED_FALSE  0x00
#define RELINKED_TRUE   0x01
#define RELINKED_NEEDED 0x02
#define RELINKED_STRUCT 0x04
#define RELINKED_BITFIELD 0x08

typedef struct ReLnkCbLstStruct
{
  TCBFCNP cbfcn;
  int lnkId;
  UINT32 cbId;
  void *cbRef;
  short hasReference;
  short isInsideCallback;
  struct ReLnkCbLstStruct *nxt;
} ReLnkCbEntry;

typedef struct ReLnkLstStruct
{
  char key[FULL_NAME_MAXSIZE+8];
  char tag[TAG_NAME_SIZE];
  char field[FIELD_NAME_SIZE];
  int  fmt;
  int  siz;
  int  lnkFmt;  /* original link format */
  BYTE *dat;    /* for possible use with structs (unbound data) */
  ReLnkCbEntry *cbLst;
  struct ReLnkLstStruct *nxt;
} ReLnkTblEntry;

extern ReLnkTblEntry *gReLnkTbl[LNKTBL_HASH_SIZE];

/**
 * \internal
 * \brief Linked List structure used by clients to manage redirected calls.
 */
typedef struct rdrLnkStruct
{
  char srcContext[CONTEXT_NAME_SIZE];
  char srcServerName[EXPORT_NAME_SIZE];
  char srcProperty[PROPERTY_NAME_SIZE];
  char srcDeviceName[DEVICE_NAME_SIZE];
  char dstContext[CONTEXT_NAME_SIZE];
  char dstServerName[EXPORT_NAME_SIZE];
  char dstDeviceName[DEVICE_NAME_SIZE];
  char dstProperty[PROPERTY_NAME_SIZE];
  struct rdrLnkStruct *prv;
  struct rdrLnkStruct *nxt;
} RdrTblEntry;
extern RdrTblEntry *rdrLnkLst;

/**
 * \internal
 * \brief Linked List structure used by connection table if multiple links to same contract
 */
typedef struct DependentLinkStruct
{
  int linkId;
  int isInactive;
  struct DependentLinkStruct *prv;
  struct DependentLinkStruct *nxt;
} DpdTblEntry;

typedef struct mcaLnkItemStruct
{
  int linkId;
  int isCanceled;
  struct mcaLnkItemStruct *nxt;
} McaLnkItem;
/**
 * \internal
 * \brief Linked List structure used by connection table if link is element of multi-channel link
 */
typedef struct mcaLinkStruct
{
  int linkId;
  int ndevices;
  int nlinks;
  McaLnkItem *links;
  char mcaContext[CONTEXT_NAME_SIZE];
  char mcaServerName[EXPORT_NAME_SIZE];
  char mcaProperty[PROPERTY_NAME_SIZE];
  char mcaDevice[DEVICE_NAME_SIZE];
  struct mcaLinkStruct *prv;
  struct mcaLinkStruct *nxt;
} McaTblEntry;
extern McaTblEntry *mcaLnkLst;
#define MCADEV_TAG "mcaDev=>"

typedef struct
{
  float absolute;
  float percent;
} DataToleranceFactor;

#define GRP_SYNC_NOSYNC 0
#define GRP_SYNC_STARTSYNC 1
#define GRP_SYNC_NOTIFICATION 2
#define GRP_SYNC_CYCLEOFFSET 3
#define GRP_SYNC_INTERVAL 4
#define GRP_SYNC_INSYNC 5

typedef struct grpMemberStruct
{
  int conTblId;
  int glbTblId;
  DTYPE dout;
  char key[STATUS_SIZE+8];
  struct grpMemberStruct *nxt;
} GrpMember;

/**
 * \internal
 * \brief Linked List structure used by clients to manage grouped calls.
 */
typedef struct grpStruct
{
  short numPending; /**< number of group events pending */
  short numInGrp; /**< number of elements in group */
  short grpCc; /**< group completion code */
  short grpSyncLevel; /**< group synchronization level */
  short grpIncMono; /**< group count increments monotonically */
  short grpHeadInc; /**< group head count has incremented */
  int grpUpdateInterval; /**< group update interval */
  int grpBndWdthT; /**< group time dispersion */
  int grpBndWdthC; /**< group cycle dispersion */
  int grpSysStamp; /**< assigned group system stamp (cycle number) */
  int grpLastStamp; /**< last group system stamp (cycle number) */
  int grpHeadLnkId; /**< designated group master's connection table id */
  struct timeval grpTimeUpdate; /**< time of this update */
  struct timeval grpLastUpdate; /**< time of last update */
  size_t grpId; /**< group identification (cbFcn address) */
  GrpMember *members; /**< group member list */
  struct grpStruct *next; /**< next entry in list */
} GrpTblEntry;

#ifdef DOS
# define WILDCARD_LINKID(i) ((i) & 0xff)
# define WILDCARD_LISTID(i) ((i) >> 8)
# define MAKE_WILDCARD_CALLBACKID(wcid,lstid) ((wcid) + ((lstid)<<8))
#else
# define WILDCARD_LINKID(i) ((i) & 0xffff)
# define WILDCARD_LISTID(i) ((i) >> 16)
# define MAKE_WILDCARD_CALLBACKID(wcid,lstid) ((wcid) + ((lstid)<<16))
#endif

/**
 * \internal
 * \brief Table entry structure used by clients to manage wildcard calls
 */
typedef struct wildCardTableStruct
{
  int linkId;  /* xref into connTableEntry */
  int cbId; /* user's callback id */
  TCBFCNP cbfcn; /* user's callback */
  int length;  /* length of device list */
  int format;  /* original caller format */
  int mode;    /* link mode */
  int numPending; /* number waiting to notify */
  int numActive;  /* number of active links in this set */
  NAME64 *list;/* list of device names */
  BYTE *data;  /* either floats or longs */
  int *status; /* list of individual comp codes */
  int *listId; /* list of individual link ids */
  short hasReference; /* user's callback prototype flag */
  short isInsideCallback;
  void *cbRef; /* user's callback reference parameter */
} WcTblEntry;

extern WcTblEntry **wcTable;
extern int nWildCardTableEntries;

typedef struct cbXRefTblStruct
{
  int linkId;  /* xref into connTableEntry */
  int cbId; /* user's callback id */
  TCBFCNP cbfcn; /* user's callback */
  short hasReference; /* user's callback prototype flag */
  short isInsideCallback;
  void *cbRef; /* user's callback reference parameter */
} cbXRefTblEntry;

extern cbXRefTblEntry **cbXRefTbl;
extern int nCbXRefTblEntries;

#define UNASSIGNED_INDEX -1
#define UNASSIGNED_CALLBACKID 0xeeeeeeee
#ifndef SMALL_TINE_LIB
# define DEFAULT_RECV_QUEUE_DEPTH 10
#else
# define DEFAULT_RECV_QUEUE_DEPTH 1
#endif
typedef struct ConnTableQueueStruct
{
  BYTE *dataOut;
  int dataSize;
  int dataStatus;
  int statusCode;
  UINT32 timeStamp;
  int timeStampMSEC;
  int usrStamp;
  int sysStamp;
  int xfrReason;
  struct ConnTableQueueStruct *next;
} ConTblQueue;

#define LWD_POLLING_RATE 1000
#define LWD_PROPERTY "SRVSTARTTIME"
#define LWD_DTAG "WDOG"
#define LWD_TIMER_DEADBAND 5
typedef struct LnkWatchDogStruct
{
  int lnkId;
  int lnkLastStatus;
  int nlinks;
  UINT32 srvStartTime;
  struct LnkWatchDogStruct *nxt;
} LnkWatchDog;
extern LnkWatchDog *wdLinks;
extern int lwdPollingInterval;
/**
 * \internal
 * \brief Connection table structure used by clients.
 */
typedef struct /* local */
{              
  char expName[EXPORT_NAME_SIZE];/**< eqm server export name */
  char prpName[PROPERTY_NAME_SIZE];/**< eqm property */
  char devName[DEVICE_NAME_SIZE];/**< eqm device name */
  char context[CONTEXT_NAME_SIZE];/**< eqm context */
  char *extStringspace; /**< eqm extended string space (long device names) */
  short extStringsize;  /**< eqm extended string space size */
  short access;         /**< data access */
  short formatIn;       /**< eqm data format down */
  short formatOut;      /**< eqm data format up */
  UINT32 sizeIn;        /**< data array size down */
  UINT32 sizeOut;       /**< data array size up */
  char strTagIn[TAG_NAME_SIZE];/**< eqm structure tag down */
  char strTagOut[TAG_NAME_SIZE];/**< eqm structure tag up */
  char *key;            /**< cross reference to ReLnkTblEntry key (if needed) */
  short dataStatus;     /**< signal new arrival (negative completion code) */
  short tocounter;      /**< link timeout counter */
  int  srvIdx;          /**< index into Srv Table */
  int  fecIdx;          /**< index into FEC database */
  UINT32 pollingInterval;/**< polling interval for connectioned clients */
  UINT32 heartbeat;     /**< used by subscriptions */
  UINT16 mode;          /**< used by subscriptions */
  UINT16 xferReason;    /**< transfer flag applicable to last data transfer */
  int tineProtocol;     /**< tine protocol level */
  int inetProtocol;     /**< transport protocol (IPX,SPX,UDP,TCP) */
  UINT32 timereq;       /**< last polling request */
  UINT32 timereqMSEC;   /**< last polling request (msec fraction) */
  UINT32 starttime;     /**< entry into table (used by subscriptions) */
  UINT32 dtimestamp;    /**< arriving TINE timestamp */
  int dtimestampMSEC;   /**< arriving TINE timestamp (msec fraction) */
  int datastamp;        /**< call applied data stamp */
  int systemstamp;      /**< system applied data stamp */
  UINT32 lastDataSize;  /**< arriving data size */
  SOCKET tcpSck;        /**< tcp/ip socket if connected */
  UINT32 mcastGrp;      /**< the multicast group ip addr (if attached to a group) */
  int  counter;         /**< current subscription counter */
  int  renewalMultiplier; /**< 'learned' renewal multiplier to establish maximum counter */
  char *dataIn;         /**< pointer to associated data down */
  char *dataOut;        /**< pointer to associated data up */
  char *bufOut;         /**< pointer to raw data up buffer if necessary */
  BYTE *bufOutBlks;     /**< pointer to data-up buffer block progress if necessary */
  char *bufCmp;         /**< pointer to comparison buffer if suppressHearbeatNotification == TRUE */
  DataToleranceFactor *dtf;/**< pointer to comparison tolerance if applied */
  short allocated;      /**< is dataOut allocated memory ? */
  short useErrValue;    /**< flag  */
  UINT16 numblks;       /**< progress counters for incoming data */
  UINT16 blksin;        /**< progress counters for incoming data */
  UINT32 bytesin;       /**< progress counters for incoming data  */
  UINT32 nextDataSize;  /**< next outgoing data size in bytes */
  UINT32 nextDataPtr;   /**< next outgoing data array pointer */
  UINT32 bytesRemaining;/**< number of bytes data remaining */
  UINT32 bytesTotal;    /**< total number of bytes data */
  UINT16 datamtu;       /**< outgoing data minimum transport unit */
  UINT16 blkidIn;       /**< block Id of data to send */
  UINT16 numblksIn;     /**< num blks input data to send */
  UINT16 blknumIn;      /**< this blk num of input data */
  UINT32 cbId;          /**< pivate callback ID */
  TCBFCNP cbfcn;        /**< callback function */
  UINT8 isGrouped;      /**< flag to indicate if element is grouped */
  UINT8 notifyPending;  /**< flag to indicate if notification is pending */
  UINT8 suppressCallback;/**< flag to withhold callback notification */
  UINT8 relinked;       /**< flag to indicate if entry is relinked (or action needed) */
  UINT16 subId;         /**< server subscription ID (applicable in MCAST) */
  UINT32 rdrId;         /**< if non-null gives child connection ID for redirected calls */
  WcTblEntry *wcLink;   /**< wildcard entry if relevant */
  short linkStatus;     /**< current link status */
  short needsNotification;/**< flag to call NotifyClient() or not */
  short hasNotifiedOnce; /**< necessary for addressing nofication suppression issues */
  short hasUserCallbackId; /**< flag : TRUE if set by caller, otherwise false */
  short lastLnkErrSrc;  /**< LCL or RMT */
  short isWildcardLink; /**< parent link is a wildcard call */
  short isWaitingSyncNotification; /**< for synchronous calls */
  short blcounter;      /**< blacklist counter */
  char lastStatusString[STATUS_SIZE]; /**< temporary storage for return status */
  BYTE  errValue[64];   /**< buffer for assigned error value */
  RdrTblEntry *rdrLink; /**< pointer into the redirection table */
  int cancelPending;    /**< delay CANCEL because of dependencies */
  int hasAcknowledged;  /**< flag to smooth subscription renewal transitions */
  int linkId;           /**< this table entry index */
  int boundToId;        /**< id of parent link if this one is bound (>0) */
  int glbTblId;         /**< globals table id if this refers to a global */
  int mcaIndex;         /**< multi-channel array index (if any) */
  McaTblEntry *mcaLink; /**< pointer to multi-channel link (if any) */
  DpdTblEntry *dpdLink; /**< pointer to dependency list (if this link is a parent) */
  int grpLatency;       /**< notification lag of this group element (if grouped) */
  int syncOffset;       /**< learned offset in sysStamp of this group element (if grouped) */
  GrpTblEntry *grp;     /**< pointer to the group table entry (if relevant) */
  int queueLen;         /**< current link queue length */
  int queueDepth;       /**< current link queue depth */
  ConTblQueue *queue;   /**< current link queue */
  short hasReference;   /**< TRUE => use protocol with cteRef ! */
  short isInsideCallback;/**< flag indicating link entry is inside callback routine */
  void *cbRef;          /**< reference pointer passed in AttachLinkEx2() */
  LnkWatchDog *wdLink;  /**< link watchdog reference if relevant */
} ConTblEntry;

typedef struct /* net */
{              
  char expName[EXPORT_NAME_SIZE]; /* eqm function name */
  char prpName[PROPERTY_NAME_SIZE]; /* eqm property */
  char devName[DEVICE_NAME_SIZE]; /* eqm device name */
  char context[CONTEXT_NAME_SIZE]; /* eqm context */
  char eqm[EQM_NAME_SIZE];
  UINT16 mode;          /* used by subscriptions */
  short access;         /* data access */
  short formatIn;       /* eqp data format down */
  short formatOut;      /* eqp data format up */
  UINT32 sizeIn;        /* data array size down */
  UINT32 sizeOut;       /* data array size up */
  char strTagIn[TAG_NAME_SIZE]; /* eqm structure tag down */
  char strTagOut[TAG_NAME_SIZE]; /* eqm structure tag up */
  UINT32 pollingInterval; /* polling rate for connectioned clients */
  UINT32 heartbeat;     /* used by subscriptions */
  UINT32 starttime;     /* entry into table (used by subscriptions) */
  UINT32 dtimestamp;    /* arriving TINE timestamp */
  int dtimestampMSEC;   /* arriving TINE timestamp (msec fraction) */
  int  counter;         /* current subscription counter */
  short linkStatus;     /* current link status */
  short reserved;
} ConTblInfo;
#define CONTBL_QUERY_SIZE sizeof(ConTblInfo)

/* GetConnectionTable */
/**
 * \brief Gets the current connection table.
 * 
 * A client's connections are managed and maintained in a connection table.  You can retrieve
 * the current connection table by using this call
 *
 * \param tbl is a reference to block of memory which will hold the ConTblInfo array
 * giving the current client connection table.
 * \param tblSize (input) is a pointer to an integer value containing the maximum size (number of entries)
 * the the tbl reference can hold. (output) will contain the size of the current connection table.
 *
 * \return 0 upon success or a TINE return code.
 *
 * \see GetConnectionTableCapacity(), SetConnectionTableCapacity()
 */
TINE_EXPORT int GetConnectionTable(ConTblInfo *tbl,int *tblSize);

#define EMPTY_PROP    0
#define TARGET_PROP   1
#define WILDCARD_PROP 2
#define STOCK_PROP    3

#ifndef PROPERTY_HASH_SIZE
# define PROPERTY_HASH_SIZE 211
#endif
#define PROPERTY_HASH_INDEX(i) ((i) & 0xff)
#define PROPERTY_NUM_INDEX(i) (((i) >> 8) & 0xffff)
#define MAKE_PROPERTY_ID(i,indx) (((i) << 8) | (indx))

#define PRP_REDIR_CTX_OFFSET 0
#define PRP_REDIR_SRV_OFFSET 32
#define PRP_REDIR_DEV_OFFSET 64
#define PRP_REDIR_PRP_OFFSET 128

#define PRP_REDIR_SRV_OFFSET_P5 0
#define PRP_REDIR_PRP_OFFSET_P5 16

#ifndef DEVICES_HASH_SIZE
# define DEVICES_HASH_SIZE 211
#endif
#define DEVICES_HASH_PAYOFF (DEVICES_HASH_SIZE/10) /* don't hash if fewer modules than this */

typedef struct HashXRef
{
  int index;
  struct HashXRef *next;
} HashXRefStruct;

#ifndef STOCKPROP_HASH_SIZE
# define STOCKPROP_HASH_SIZE 103
#endif

extern HashXRefStruct **StockPropXRefTable;

typedef struct 
{
  NAME64 *prpList;
  int prpListSize;
  char filename[64]; /* used as a 'tag' */
} EqmDevicePropertyList;

/**
 * \internal
 * \brief export info struct (to handle file input)
 *
 */
typedef struct ExpInfoStruct
{
  char EqmContext[CONTEXT_NAME_SIZE]; /* exported equipment name context */
  char ExpName[EXPORT_NAME_SIZE]; /* exported equipment module name */
  char SubSystem[SUBSYSTEM_NAME_SIZE]; /* equipment module sub system */
  char EqmName[EQM_NAME_SIZE]; /* local equipment module name */
  char EqmGroup[EXPORT_NAME_SIZE]; /* group name, if applicable */
  char EqmMaster[EXPORT_NAME_SIZE]; /* eqm master name, if applicable */
  char EqmSlaveMaster[EXPORT_NAME_SIZE]; /* eqm master name for slave operation, if applicable */
  char PrpName[PROPERTY_NAME_SIZE];
  char fmtOut[64];
  char fmtIn[64];
  char description[128];
  char redirection[DEVICE_REDIR_SIZE];
  char access[ACCESS_STR_SIZE];
  int  PrpSizeOut;
  int  PrpSizeIn;
  int  PrpId;
  int  EqmGroupIdx;        /* group index if applicable */
  int  DevSpace;/* total amount of space initially allocated */
} XIS;

typedef struct ContractResourceStruct
{
  DTYPE din;
  DTYPE dout;
  char devName[DEVICE_NAME_MAXSIZE];
  char devProp[PROPERTY_NAME_SIZE+4];
  short access;
} CRS;

typedef struct 
{
  char *rdrName;
} EqmDeviceRdrListType;

/**
 * \internal
 * \brief callback and dispatch function prototypes (normal and extended)
 *
 */
typedef void (*SYSTSKP)(void);
typedef int (*EQMFCNP)(char *,char *,DTYPE *,DTYPE *,short);
typedef int (*XEQMFCNP)(char *,char *,DTYPE *,DTYPE *,short,void *);
typedef void (*EQMTSKP)(void);
typedef void (*XEQMTSKP)(void *);

#define EQMCALL(el,dn,pn,od,id,ax) (\
  el->hasReference ?\
  ((XEQMFCNP)el->eqmFcn)(dn,pn,od,id,ax,el->EqmRef) :\
  ((EQMFCNP)el->eqmFcn)(dn,pn,od,id,ax)\
)

#define TSKCALL(el,tsk) (\
  el->hasReference ?\
  ((XEQMTSKP)el->tsk)(el->EqmRef) :\
  ((EQMTSKP)el->tsk)()\
)

/**
 * \internal
 * \brief Client struct for protocol level 5.
 *
 */
typedef struct /* struct is passed over the net (up to the transport field) */
{
  char               userName[USERNAME_SIZE];
  IPXAddress         IPXaddress;  /* if 0 then IP transfer is assumed */
  BYTE               node[6];     /* IPX routing node (Immediate Address) */
  struct sockaddr_in IPaddress;
  short              ncontracts;  /* number of registrations */
  short              tineProtocol;/* target client protocol level */
  short              inetProtocol;/* transport protocol */
  SOCKET             sck;         /* communication socket */
} ClnHdr;
#define CLIENT_SIZE (USERNAME_SIZE + 12 + 6 + 16 + 2 + 2 + 2 + sizeof(int))

/**
 * \internal
 * \brief Definitions relevant to an Access Lock
 *
 */
#define LOCK_UNLOCKED 0
#define LOCK_PREEMPTIVE 1
#define LOCK_PERSISTENT 2
#define LOCK_ABORT 3
#define LOCK_XREAD 0x100
#define MAX_LOCK_DURATION 60

/**
 * \internal
 * \brief AccessLock used on the server side
 *
 */
typedef struct
{
  ClnHdr cln;
  int lockType;
  int lockDuration;
  UINT32 lockToken;
  UINT32 lockStart;
} AccessLock;

/**
 * \internal
 * \brief AccessLock used on the client side (link-listed)
 *
 */
typedef struct AccessLockListItemStruct
{
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  int lockType;
  int lockDuration;
  int lockLinkId;
  int lockLinkStatus;
  int lastSent;
  struct AccessLockListItemStruct *nxt;
} AccessLockListItem;

extern AccessLockListItem *accessLockLst;
int rmvAccessLock(char *context, char *server,int wait);
int _setAccessLock(char *context, char *server, int lockType,int lockDuration);
AccessLockListItem *getAccessLockItem(char *context, char *server);
void checkAccessLockItems(void);
int GetDeviceAclInfo(char *eqm,int devnr);

typedef struct 
{
  NAME16 *usersAllowed;     /* dev-specific write access list */
  int numUsersAllowed;      /* number of allowed users */
  struct sockaddr_in *netsAllowed; /* dev-specific write access list */
  int numNetsAllowed;       /* number of allowed ip nets */
} DeviceSecurityStruct;

typedef struct
{
  char grpName[DEVICE_NAME_SIZE];
  int grpSize;
  int grpIndex;
} DeviceMcaGroup;

typedef struct
{
  NAME64 name;
  char *desc; /* device description (if any) */
  char *rdrName; /* redirection string (if any) */
  EqmDevicePropertyList *prpLst;/* ptr to device-property list (if any) */
  DeviceSecurityStruct *aclLst; /* ptr to device security list (if any) */
  DeviceMcaGroup *mcaDev; /* ptr to specific MCA group for this device (if any) */
} ExportDeviceListStruct;

/**
 * \internal
 * \brief Linked list structure used to hold equipment module information.
 *
 */
typedef struct ExportListTag /* local */
{
  char   EqmContext[CONTEXT_NAME_SIZE]; /* exported equipment name context */
  char   EqmExportName[EXPORT_NAME_SIZE]; /* exported equipment module name */
  char   EqmSubSystem[SUBSYSTEM_NAME_SIZE]; /* equipment module sub system */
  char   EqmName[EQM_NAME_SIZE]; /* local equipment module name */
  char   EqmAlias[EQM_NAME_SIZE]; /* local equipment alias name */
  char   EqmGroup[EXPORT_NAME_SIZE]; /* group name, if applicable */
  int    groupIndex;        /* group index if applicable */
  int    exportId;          /* local ID */
  ExportPropertyListStruct  /* pointer to chained equipment property hash list */
         *EqmPrpLst[PROPERTY_HASH_SIZE+1]; /* Add room for Wild-card list */
  int    (*EqmPropertyQueryFcn)(char *devName,PrpQueryStruct **prpqs);
  int    (*XEqmPropertyQueryFcn)(char *devName,XPropertyQueryStruct **xpqs);
  short   hasPropertyDeviceNames;
  short   hasRegisteredDevicesNames;
  int    EqmWcPrpLstCapacity;/* size of list buffer space used in eqm wildcard queries */
  NAME64 *EqmWcPrpLst;      /* pointer to property list buffer space used in eqm wildcard queries */
  NAME64 *EqmDevLst;        /* pointer to device name list */
  NAME64 *EqmDevDescLst;    /* pointer to device description list */
  EqmDeviceRdrListType *EqmDevRdrLst; /* pointer to device redirection list (list of redirection strings) */
  EqmDevicePropertyList *EqmDevPrpLst;/* pointer to array of device-property lists */
  DeviceSecurityStruct *EqmDevAclLst; /* pointer to array of 'DeviceSecurityStruct's */
  DeviceMcaGroup *EqmDevMcaLst; /* pointer to array of device MCA group elements */
  HashXRefStruct **EqmDevXRef; /* ModuleName cross-reference (DEVICES_HASH_SIZE in length) */
  short  EqmNumProperties;  /* total number of properties supported */
  short  EqmNumDevices;     /* total number of devices supported */
  short  EqmDevLstCapacity; /* total amount of space initially allocated */
  short  registered;        /* registered with name server */
  ADSList *adsTable;        /* alarm definition structure table (linked List) */
  short  adsTableSize;      /* number of alarm definitions */
  short  alarmSystem;       /* alarm system flag */
  ALARM  **almLst;          /* associated alarm list */
  AMSDynSet amsDynSet;      /* brief alarm message containing totals */
  short  nalarms;           /* number of active alarms */
  UINT32 almMRT;            /* alarm Most Recent Timestamp */
  UINT32 almMRCT;           /* alarm Most Recent Complete Timestamp */
  short  almHSV;            /* alarm highest Severity */
  short  nAlmMRT;           /* number of alarms at MRT */
  short  nAlmMRCT;          /* number of alarms at MRCT */
  short  nAlmHSV;           /* number of alarms at HSV */
  short  nclients;          /* number of active clients */
  short  ncontracts;        /* number of active contracts */
  EQMFCNP eqmFcn;           /* associated equipment function */
  EQMTSKP eqmBkg;           /* associated background task */
  THRHANDLE hEqmBkg;        /* handle to background task (in case of threads) */
  SINT32 rateBkg;           /* background polling rate msec */
  short bkgbusy;            /* background busy semaphore */
  short inidone;            /* init task done flag */
  short hstdone;            /* history init done flag */
  short almdone;            /* alarm init done flag */
  short isidle;             /* eqm + tasks are in an idle state */
  short isbusy;             /* eqm fcn call is busy */
  short busycnt;            /* busy counter for 'not_ready' */
  short isterminating;      /* termination signal */
  struct timeval lastBkg;   /* background task last polled */
  EQMTSKP eqmIni;           /* associated initialization  */
  EQMTSKP eqmExi;           /* associated cleanup */
  char EqmLastErrorString[STATUS_SIZE];
  int EqmLastErrorCode;
# ifdef MULTITHREADED
  CRS EqmCrs;               /* ContractResourceStruct (mutex protected) */
# ifdef UNIX
  pthread_mutex_t EqmBusyMutex;
  pthread_mutex_t EqmCallMutex[3];
  pthread_mutex_t DataProtectionMutex;
# endif
  MXHANDLE hEqmBusyMutex;
  MXHANDLE hEqmCallMutexTriad[3];
  MXHANDLE hDataProtectionMutex;
  THRHANDLE hEqmCallThread;
# else
  int hEqmBusyMutex;
# endif
  void *EqmCls;             /* contract list structure reference */
  int hasReference;         /* TRUE => use prototypes with EqmRef ! */
  void *EqmRef;             /* reference pointer to use in extended dispatch functions */
  time_t eqmStartupTime;    /* eqm specific startup flag */
  int contractId;           /* last contract id */
  char usersFileName[32];   /* either users.csv or <eqm>/users.csv */
  NAME16 *usersAllowed;     /* eqm-specific write access list */
  int numUsersAllowed;      /* number of allowed users */
  int checkUserAccess;      /* check users ACL : TRUE or FALSE */
  int usersScanFlag;        /* -1 => don't scan, 1 => scan in FEC_HOME, 2 => scan in <eqm> dir */
  char netsFileName[32];    /* either ipnets.csv or <eqm>/ipnets.csv */
  struct sockaddr_in *netsAllowed; /* eqm-specific write access list */
  int numNetsAllowed;       /* number of allowed ip nets */
  int checkNetsAccess;      /* check users ACL : TRUE or FALSE */
  int netsScanFlag;         /* -1 => don't scan, 1 => scan in FEC_HOME, 2 => scan in <eqm> dir */
  int failoverType;         /* _NONE, _MASTER, or _SLAVE */
  char EqmMaster[EXPORT_NAME_SIZE]; /* applicable if master or slave */
  char EqmSlaveMaster[EXPORT_NAME_SIZE]; /* applicable if slave */
  AccessLock accessLock;    /* possible access lock */
  struct ExportListTag *next; /* Linked list */
} ExportListStruct;

extern int gEqmFcnAccessor;
extern ExportPropertyListStruct *ExportPropertyList;
extern ExportListStruct *ExportList;
extern int gEqmListSize;
extern int gSystemUseDataProtection;

/**
 * \internal
 * \brief A Structure which is used to hold save-and-restore information at initialization
 * 
 * run thru during initial pass of serverCycle()
 */
typedef struct srLstStruct
{
  ExportListStruct *el;
  ExportPropertyListStruct *pl;
  int restored;
  struct srLstStruct *nxt;
} srLst;
extern srLst *gSrLst;
void applyStoredPrpValues(ExportListStruct *el);
void writeSrValues(ExportListStruct *el,char *dev,char *prp,DTYPE *din);
void addSrPrp(ExportListStruct *el,ExportPropertyListStruct *pl);
void freeSrPrpList(void);
char *getDevNameAtIndex(ExportListStruct *el,ExportPropertyListStruct *pl,int idx);


/**
 * \internal
 * \brief A Structure which is used to map a EqmContext/ExportName to a FecName/EqmName.
 * 
 * The client uses EqmContext and Export name to address the FEC. 
 * TINE internally uses the FecName and the Equipment name.
 * 
 * The TINE library uses this structure for internal caches.
 */
typedef struct                      /* struct is passed over the net (old, no swapping) */
{
  char  FecName[FEC_NAME_SIZE];     /**< FEC Name */
  char  SubSystem[SUBSYSTEM_NAME_SHORTSIZE]; /* short sub-system name */
  char  ExportName[EXPORT_NAME_SHORTSIZE];	/**< Export Name */
  char  EqmName[EQM_NAME_SHORTSIZE+1];	  /**< Equipment name */
  char  EqmContext[CONTEXT_NAME_SHORTSIZE];	/**< EqmContext */
  short alarmSystem;
} RPCDataStruct;
#define RPCDATASTRUCT_SIZE (FEC_NAME_SIZE+SUBSYSTEM_NAME_SHORTSIZE+EXPORT_NAME_SHORTSIZE+EQM_NAME_SHORTSIZE+1+CONTEXT_NAME_SHORTSIZE+2)

typedef struct                      /* struct is passed over the net (new) */
{
  char  FecName[FEC_NAME_SIZE];
  char  SubSystem[SUBSYSTEM_NAME_SIZE];
  char  ExportName[EXPORT_NAME_SIZE];
  char  EqmName[EQM_NAME_SIZE];
  char  EqmContext[CONTEXT_NAME_SIZE];
} ExpDataStruct;
#define EXPDATASTRUCT_SIZE (FEC_NAME_SIZE+SUBSYSTEM_NAME_SIZE+EXPORT_NAME_SIZE+EQM_NAME_SIZE+CONTEXT_NAME_SIZE)

extern UINT32 IncomingNetworkProtocol;
typedef struct                      /* struct is passed over the net (old, no swapping) */
{
  char fecName[FEC_NAME_SIZE];
  BYTE reserved[3];
  BYTE Port[2];
  BYTE Network[4];
  BYTE Node[6];
  BYTE ImmediateAddress[6];
  char IP[16];
  BYTE portOffset;
  char IPh_addr[4];
  short  Prot;
} RPCFecStruct;
#define RPCFECSTRUCT_SIZE (FEC_NAME_SIZE+3+2+16+21+2)

/* these macro re-definitions of the IPX address fields will make sections of the code easier to read */
#define MMFhwnd IPXImmediateAddress
#define PIPEsck IPXNetwork
#define PIPEerr IPXNode[0]
typedef struct                      /* struct is passed over the net (new) */
{
  char fecName[FEC_NAME_SIZE];
  BYTE IPXNetwork[4];
  BYTE IPXNode[6];
  BYTE IPXImmediateAddress[6];
  char IP[ADDR_SIZE];
  char IPh_addr[4];
  SINT32 portOffset;
  SINT32 inetProtocol;
  SINT32 tineProtocol;
} FecDataStruct;
#define FECDATASTRUCT_SIZE (FEC_NAME_SIZE+16+16+4+3*4)

typedef struct
{
  char os[12];
  char desc[64];
  char loc[32];
  char hdw[32];
  char sub[16];
  char resp[32];
} RPCFecExStruct;

typedef struct /* new */
{
  char os[FEC_OS_SIZE];
  char desc[FEC_DESC_SIZE];
  char loc[FEC_LOCATION_SIZE];
  char ver[FEC_VERSION_SIZE];
  char hdw[FEC_HDW_SIZE];
  char resp[FEC_RESP_SIZE];
} FecInfoStruct;

typedef struct /* used for holding data temporarily during initialization */
{
  char fecName[FEC_NAME_SIZE];
  char fecOs[FEC_OS_SIZE];
  char fecContext[CONTEXT_NAME_SIZE];
  char fecSubsystem[SUBSYSTEM_NAME_SIZE];
  char fecDescription[FEC_DESC_SIZE];
  char fecLocation[FEC_LOCATION_SIZE];
  char fecHardware[FEC_HDW_SIZE];
  char responsible[FEC_RESP_SIZE];
  char expNameTarget[EXPORT_NAME_SIZE];
  char eqmNameTarget[EQM_NAME_SIZE];
  char historyHome[HISTORY_HOME_SIZE];
  int  fecPort;
} FIDS;

typedef struct 
{
  int  devnum;
  char devname[DEVICE_NAME_SIZE];
  char devdesc[DEVICE_DESC_SIZE];
  char devrdr[DEVICE_REDIR_SIZE];
  char prplst[32];
} DEVIS;

typedef struct
{
  char FecName[FEC_NAME_SIZE];
  char reserved[4];
  UINT32 localtime;
  UINT32 starttime;
  UINT32 systemPollingRate;
  short  numBkgTsks;
  short  numTotalContracts;
  short  numTotalClients;
  short  numTargetContracts;
  short  numTargetClients;
  short  numConnections;
  UINT32 numConnectionTimeouts;
  UINT32 numConnectionArrivals;
  UINT32 numUDPpkts;
  UINT32 numTCPpkts;
  UINT32 numIPXpkts;
  UINT32 numSPXpkts;
} ActivityQueryStruct;
#define ACTIVITYQUERYSTRUCT_SIZE (FEC_NAME_SIZE+4+4*3+2*6+4*6)

typedef struct
{
  char desc[64];
  BYTE dataFmt;
  BYTE dataSiz;
  BYTE data[64];
  UINT32 timeEntered;
  UINT32 codeNow;
  UINT32 codePrv;
  UINT32 codeNxt;
} ServerStateStruct;
#define SERVERSTATESTRUCT_SIZE (64+2+64+4*4)

typedef struct
{
  UINT32 AveBusyTime;
  UINT32 CycleCounts;
  UINT32 MaxCycleCounts;
  UINT32 SingleLinkCount;
  UINT32 ClientMisses;
  UINT32 ClientReconnects;
  UINT32 ClientRetries;
  UINT32 ContractMisses;
  UINT32 ContractDelays;
  UINT32 BurstLimitReachedCount;
  UINT32 DataTimeStampOffset;
} ServerStatsStruct;
#define SERVERSTATSSTRUCT_SIZE (11*4)
#ifndef DOS
typedef struct 
{
  UINT32 srvWorkArea; /* srvWorkAreaSize */
  UINT32 cycDeadBand; /* gSystemCycleDeadband */
  UINT32 minPollInterval; /* MinPollingRate */
  UINT32 conTblCapacity; /* ContractListCapacity */
  UINT32 conRenewalLen; /* gSubRenewalLength */
  UINT32 clnTblCapacity; /* ClientListCapacity */
  UINT32 ackOnChange; /* RequireAcknowledgments */
  UINT32 srvBurstLimit; /* gBurstLimit */
  UINT32 srvBurstCycDly; /* gCycleDelay */
  UINT32 srvPacketMTU; /* gPacketMtu */
  UINT32 tcpMaxMsgSize; /* gMaxTcpMsgSize */
  UINT32 tcpTblCapacity; /* tcpConTblSize */
  UINT32 srvSendBuffers; /* max_srvrcvbuf */
  UINT32 srvRecvBuffers; /* max_srvsndbuf */
  UINT32 srvLazySched; /* gSchedulePropertyLazy */
  UINT32 srvBkgTskRentr; /* gAllowBkgTaskReentrancy */
  UINT32 srvCycInSepThrd; /* gRunServerCycleInSeparateThread */
  UINT32 tcpTraInSepThrd; /* gRunTransportInSeparateThread */
  UINT32 clnWorkArea; /* clnWorkAreaSize */
  UINT32 lnkTblCapacity; /* ConTblCapacity */
  UINT32 clnBurstLimit; /* gClientBurstLimit */
  UINT32 clnSendBuffers; /* max_clnrcvbuf */
  UINT32 clnRecvBuffers; /* max_clnsndbuf */
  UINT32 clnRecvQueue; /* gClnRecvQueueDepth */
  UINT32 useLoopbackAddr; /* gUseLoopback */
  UINT32 useWdogLinks; /* gAutoLinkWatchdogs */
  UINT32 allowCommonLnks; /* allowCommonLinks */
  UINT32 retryOnTimeout; /* gAlwaysRetry */
  UINT32 srvSchedIntvl; /*gSchedulerInterval; */
  UINT32 srvRetrdRemovl; /* gRetardSingleContractRemoval */
  UINT32 srvAllowRemote; /* gAllowRemoteExit, etc. */
  UINT32 reserved4; 
} ServerSettingsStruct;
#define SERVERSETTINGSSTRUCT_SIZE (32*4)
#endif
typedef struct
{
  char user[USERNAME_SIZE];
  char addr[ADDR_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  double timestamp;
} WriteAccessQuery;
#define WRITEACCESS_QUERY_SIZE sizeof(WriteAccessQuery)
typedef struct
{
  char user[USERNAME_SIZE];
  char addr[ADDR_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  double timestamp;
  char eqm[EQM_NAME_SIZE];
} WriteAccessInfo;
#define WRITEACCESS_STRUCT_SIZE sizeof(WriteAccessInfo)
#define WRITEACCESS_TBL_SIZE 100
extern WriteAccessInfo wrAccTbl[WRITEACCESS_TBL_SIZE];
extern int wrAccTblPtr;

/**
 * \internal
 * \brief Determines the default value of MaxNumClients
 *
 * The places an upper limit on the number of simultaneous clients a server is willing
 * to manage.  If it is known that a server needs to service more clients than this
 * default, then the value of MaxNumClients should be set accordingly at initialization time.
 */
#ifndef CLIENTLIST_CAPACITY
# define CLIENTLIST_CAPACITY     100
#endif

/**
 * \internal
 * \brief Determines the default value of MaxNumContracts
 *
 * The places an upper limit on the number of simultaneous contracts a server is willing
 * to manage.  If it is known that a server needs to service more contracts than this
 * default, then the value of MaxNumContracts should be set accordingly at initialization time.
 */
#ifndef CONTRACTLIST_CAPACITY
# define CONTRACTLIST_CAPACITY     1000
#endif

#define CTR_RENEWAL     60
#define CTR_REMINDER    5
#define CTR_ACK_REQUEST (2*CTR_REMINDER)
#define CTR_MAXIMUM     32767
#define HBT_NONE        0
#define HBT_MINIMUM     1
#define HBT_REPLY_PENDING 3
#define HBT_MAXIMUM     60
#define HBT_SUB_PREPENDING 0x3ffe
#define HBT_SUB_PENDING 0x3fff
#define HBT_ACK_PENDING 0x7fff

/**
 * \internal
 * \brief The payload size of an IPX Datagram.
 *
 */
#define IPX_DATA_SIZE   512
#define IPX_BUFFER_SIZE 1466

/**
 * \internal
 * \brief The payload size of a UDP Datagram.
 *
 * The value here works almost everywhere except NIOS (where #define UDP_BUFFER_SIZE is
 * set to 1024 in project.def).
 *
 * \note TINE performs its own packet reassembly for incoming datagrams.  For many OSes 
 * this is not necessary, but as packet reassembly cannot be assumed to be present or 
 * might have an upper limit relating to the number of recv buffers, it is a fundemental
 * feature of TINE.  
 */
#ifndef UDP_BUFFER_SIZE
# define UDP_BUFFER_SIZE 1472
#endif

#ifndef UDP_MAXBUFFER_SIZE
# define UDP_MAXBUFFER_SIZE 0x10000
#endif

#ifndef MMF_BUFFER_SIZE
# define MMF_BUFFER_SIZE 65000
#endif

#ifndef TCP_BUFFER_SIZE
# define TCP_BUFFER_SIZE 1024*16
#endif

#define REQ_BUFFER_SIZE 2000

/**
 * \internal
 * \brief The default value for one of srvWorkAreaSize, WorkAreaSize, and/or tmpWorkAreaSize.
 *
 * TINE uses a work area buffer for managing contracts calls.  On some (single-threaded) platforms 
 * all work areas point to the same buffer.
 */
#ifndef WORK_SIZE
# define WORK_SIZE  0x10000 /* 64 KB */
#endif

/**
 * \internal
 * \brief The default value for one of clnWorkAreaSize.
 *
 * TINE uses a work area buffer for managing client calls.  On some (single-threaded) platforms 
 * this points to the same area as the srvWorkAreaSize.
 */
#ifndef MIN_WORK_SIZE
# define MIN_WORK_SIZE 0x10000
#endif

/**
 * \internal
 * \brief The maximum payload size for incoming network globals.
 *
 */
#define GCAST_BUFFER_SIZE IPX_BUFFER_SIZE

/**
 * \internal
 * \brief The buffer size for a console command
 *
 */
#ifndef CMDSIZE
# define CMDSIZE 256
#endif

extern int nconsumers;
extern int ncontracts;
extern int StaleData;
extern int nExpiredContracts;

#define isWildcardContract(s) (strchr(s,'*') != NULL)
int EqmFcnWithWildcardDevice(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access);
int EqmFcnWithWildcardProperty(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access);

/**
 * \internal
 * \brief Contract input header for protocol level 5.
 *
 */
typedef struct             /* struct is passed over the net  */
{
  char   EqmProperty[PROPERTY_NAME_SIZE];
  char   EqmDeviceName[DEVICE_NAME_SIZE];
  char   EqmName[EQM_NAME_SIZE];
  UINT32 EqmSizeIn;
  UINT32 EqmSizeOut;
  BYTE   hEqmName;
  BYTE   EqmAccess;
  BYTE   EqmFormatIn;
  BYTE   EqmFormatOut;
  char   strTagIn[TAG_NAME_SIZE];
  char   strTagOut[TAG_NAME_SIZE];
} CONTRACT;
#define CONTRACT_SIZE (PROPERTY_NAME_SIZE+DEVICE_NAME_SIZE+EQM_NAME_SIZE+4*2+4*1+2*TAG_NAME_SIZE)

typedef struct
{
  CONTRACT contract;
  SINT32 pollingInterval;
  SINT32 nclients;
} ContractQueryStruct;
#define CONTRACT_QUERY_SIZE (CONTRACT_SIZE + 4 + 4)

/* tineProtocol level 5  : TINE 3.0 - 3.31 */
typedef struct             /* struct is passed over the net  */
{
  char   EqmProperty[PROPERTY_NAME_SHORTSIZE];
  char   EqmDeviceName[DEVICE_NAME_SHORTSIZE];
  char   EqmName[EQM_NAME_SIZE];
  UINT32 EqmSizeIn;
  UINT32 EqmSizeOut;
  BYTE   hEqmName;
  BYTE   EqmAccess;
  BYTE   EqmFormatIn;
  BYTE   EqmFormatOut;
  char   strTagIn[TAG_NAME_SHORTSIZE];
  char   strTagOut[TAG_NAME_SHORTSIZE];
} CONTRACT5;
#define CONTRACT5_SIZE (PROPERTY_NAME_SHORTSIZE+DEVICE_NAME_SHORTSIZE+EQM_NAME_SIZE+4*2+4*1+2*TAG_NAME_SHORTSIZE)

typedef struct
{
  CONTRACT5 contract;
  SINT32 pollingInterval;
  SINT32 nclients;
} Contract5QueryStruct;
#define CONTRACT5_QUERY_SIZE (CONTRACT5_SIZE + 4 + 4)

/**
 * \internal
 * \brief Client query struct for protocol level 5.
 *
 */
typedef struct /* struct is passed over the net */
{
  char               userName[USERNAME_SIZE];
  IPXAddress         IPXaddress;  /* if 0 then IP transfer is assumed */
  BYTE               node[6];     /* IPX routing node (Immediate Address) */
  struct sockaddr_in IPaddress;
  short              ncontracts;  /* number of registrations */
  short              tineProtocol;/* target client protocol level */
} CLNQS;

typedef struct
{
  CLNQS cln;
} ClientQueryStruct;
#define CLIENT_QUERY_SIZE (USERNAME_SIZE + 12 + 6 + 16 + 2 + 2)

/**
 * \internal
 * \brief Client connection status struct used by TINE Engine.
 *
 */
typedef struct                /* local */
{
  short  id;                  /* contract ID as seen by consumer */
  short  mode;                /* SINGLE,DATACHANGE,TIMER,CANCEL */
  SINT32 pollingInterval;     /* requested polling interval */
  struct timeval lasttime;    /* time of most recent update */
  UINT32 starttime;           /* start of initial subscription */
  UINT16 counter;             /* counts remaining in subscription(s) */
  UINT16 Stale;               /* > 0 => send new data to consumer */
  UINT16 numblks;             /* total number of message blocks */
  UINT16 blknum;              /* block number of this message */
  short  statusCode;          /* interim completion */
  UINT16 mtu;                 /* maximal transport unit */
  UINT32 nextDataSize;        /* next outgoing data size in bytes */
  UINT32 nextDataPtr;         /* next outgoing data array pointer */
  UINT32 bytesRemaining;      /* number of bytes data remaining */
  UINT32 bytesTotal;          /* total number of bytes data */
  UINT32 misses;              /* data return after expiration */
  UINT32 reconnects;          /* client unexpectedly re-establishes link */
  int    isRetry;             /* the contract is being resubmitted by the client */
} ClnSts;

#define MAXLONGGRAMS 5
typedef struct
{
  UINT32 haddr;               /* identify client by IP address */
  UINT16 subId;               /* supplied by the consumer upon registration */
  UINT16 blksin;              /* # blocks arrived from long datagram */
  UINT16 blkid;               /* block id of long datagram */
  UINT8  busy,
         reserved;
  UINT8  *data;
  UINT32 blktime;
  UINT32 SubsStarttime;
} ContractProgress;

typedef struct ClnLstStruct
{
  ClnHdr *client;
  ClnSts status;
  struct ClnLstStruct *next;
} ClnLst;

/**
 * \brief Client Information Structure used in GetCallerInformation.
 *
 */
typedef struct ClnInfoStruct
{
  short  accessMode; /**< this client's requested access mode: CM_SINGLE, CM_DATACHANGE, CM_TIMER, etc. */
  SINT32 pollingRate;/**< this client's requested polling rate */
  UINT32 starttime;  /**< this client's start of initial subscription */
  UINT16 counter;    /**< this client's counts remaining in subscription(s) */
  char userName[USERNAME_SIZE]; /**< this client's user name */
  struct sockaddr_in addr; /**< this client's ip address */
  short ncontracts;  /**< this client's total number of contracts */
  short tineProtocol;/**< this client's tine protocol level */
  short inetProtocol;/**< this client's transport protocol: UDP, TCP, STREAM , etc.  */
} ClnInfo;

#define CONTRACT_ACTIVE    0x00
#define CONTRACT_EXPIRED   0x01
#define CONTRACT_AVAILABLE 0x02

/**
 * \internal
 * \brief Contract List Table.
 *
 */
typedef struct                  /* local */
{
  ClnLst   *clnLst;             /* assciated clnLst list */
  short    nclients;            /* number of consumers for this contract */
  short    contractID;          /* local contract ID */
  CONTRACT contract;            /* see above */
  char     *extStringSpace;     /* extended string buffer */
  UINT32   extStringSize;       /* extended string size */
  char     *lngDeviceName;      /* long device name buffer */
  char     *lngPropertyName;    /* long property name buffer */
  SINT32   pollingInterval;     /* in mseconds, smallest number wins ... */
  struct   timeval lasttime;    /* most recent time the data polled */
  struct   timeval lastmarked;  /* last time the data were marked STALE */
  struct   timeval starttime;   /* start time of the current call */
  UINT32   dtimestamp;          /* data timestamp */
  int      dtimestampMSEC;      /* data timestamp (msec fraction) */
  int      datastamp;           /* user data stamp */
  int      sysstamp;            /* system data stamp */
  BYTE     *dataOut;            /* data to send  */
  BYTE     *dataIn;             /* data received */
  BYTE     *dataBuf;            /* data working area */
  BYTE     *dataOutArr;         /* data working array for CF_STRING output */
  DTYPE    *dout;               /* prepared output DTYPE */
  DTYPE    *din;                /* prepared input DTYPE */
  UINT32   dataOutCapacity;     /* size of dataOut allocation */
  UINT32   dataSizeOut;         /* output data size < contract.dataSize */
  int      isStkCall;           /* is a stock-property call ? */
  int      isCalledInSeparateThread;
  int      callPending;         /* eqm or stk call is pending (multi-threaded?) */
  int      deliveryPending;     /* pending flag for doDelivery() vs doScheduler() */
  int      expired;             /* expired flag for contract removal */
  int      compStatus;          /* completion status */
  int      returnCode;          /* contract completion return code */
  UINT32   delays;              /* delayed scheduling due to deliveryPending */
  UINT32   misses;              /* missed scheduling (i.e. > 2x polling rate) */
  char     compString[STATUS_SIZE];/* completion string (redirection ?) */
  ExportListStruct *exp;        /* associated export element */
  ExportPropertyListStruct *prp;/* associated property element */
  int      listId;              /* cross-reference to this entry */
  int      unbuffered;          /* payload data are unbuffered : dataOut = dataBuf */
  int      accessFlags;         /* additional access flags (CA_REPEAT, CA_STATIC) */
  int      signalPending;       /* needs to call a signal handler */
} ContractListStruct;

/**
 * \internal
 * \brief Contract returned header struct
 *
 */
typedef struct                  /* struct is passed over the net : tineProtocol 5 (out) */
{
  UINT16 msgsize;               /* message size in bytes */
  UINT16 subId;                 /* supplied by the consumer upon registration */
  UINT16 CompletionCode;        /* completion or interim status */
  UINT16 numblks;               /* total number of message blocks */
  UINT16 blknum;                /* block number of this message */
  UINT16 mtu;                   /* maximum transport unit */
  UINT16 timestampMSEC;         /* msec fraction of data timestamp */
  BYTE   EqmFormat;             /* Eqp Format of returned data */
  BYTE   counter;               /* tell consumer how much is left */
  UINT32 timestamp;             /* data timestamp */
  UINT32 ClnStarttime;          /* supplied by the consumer upon registration (@ byte 20)*/
} PrdrHdr5;
#define PHDR5_SIZE (7*2+2*1+2*4)

#define CX_NULL      0x00
#define CX_RESPONSE  0x01
#define CX_STALE     0x02
#define CX_HEARTBEAT 0x04
#define CX_EVENT     0x08
#define CX_TIMER     0x10
#define CX_MESSAGE   0x20
#define CX_ERROR     0x40
#define CX_ATTACH    0x80
#define CX_RETRY    0x100

extern int gPacketMtu;
#define NPMTU_OLDWAY(np,m) ((np) == IPX ? IPX_BUFFER_SIZE :\
                     (np) == MMF ? MMF_BUFFER_SIZE :\
                     ((np) == TCP && (m) > UDP_BUFFER_SIZE) ? TCP_BUFFER_SIZE : UDP_BUFFER_SIZE)
#define NPMTU(np,m) ((np) == IPX ? IPX_BUFFER_SIZE :\
                     (np) == MMF ? MMF_BUFFER_SIZE :\
                     (np) == TCP ? TCP_BUFFER_SIZE : UDP_BUFFER_SIZE)
#define DGMTU(m,np,tp) ((np) != UDP || (tp) == 5 || (m) == 0 ? NPMTU(np,m) : (m))
/* #define DGMTU(m,np,tp) ((m) > 0 ? (m) : NPMTU(np)) */

typedef struct                  /* struct is passed over the net : tineProtocol 6 (out) */
{
  UINT16 msgsize;               /* message size in bytes */
  UINT16 subId;                 /* supplied by the consumer upon registration */
  UINT16 CompletionCode;        /* return completion code from the eqm */
  UINT16 numblks;               /* total number of message blocks */
  UINT16 blknum;                /* block number of this message */
  UINT16 mtu;                   /* maximum transport unit */
  UINT16 EqmFormat;             /* Eqp Format of returned data */
  UINT16 counter;               /* tell consumer how much is left */
  UINT16 tineProtocol;          /* tine protocol level */
  UINT16 xferReason;            /* transfer reason flag (one of the CX_ codes above) */
  UINT32 ClnStarttime;          /* supplied by the consumer upon registration (@ byte 20)*/
  UINT16 stssize;               /* size of error/status string space */
  UINT16 stscode;               /* access status code */
  UINT32 timestamp;             /* data timestamp */
  UINT32 timestampUSec;         /* usec fraction of data timestamp */
  UINT32 userstamp;             /* application settable data stamp */
  UINT32 sysstamp;              /* systematic data stamp */
} PrdrHdr;
#define PHDR_SIZE (12*2+5*4)
#define PMTU(m,np,tp)  (DGMTU(m,np,tp) - ((tp) == 5 ? PHDR5_SIZE : PHDR_SIZE) - sizeof(UINT16))
#define PHDR_STTM_OFFSET 20
#define PHDR_PROT_OFFSET 16
#define PHDR_STSSIZ_OFFSET 24
#define PHDR_SUBID_OFFSET 2
#define PHDR_CC_OFFSET 4
#define PHDR_FMT_OFFSET 12
#define PDHR_SYST_OFFSET 40
#define PDHR_URST_OFFSET 36

/**
 * \internal
 * \brief Contract header struct
 *
 */
typedef struct subscription  /* struct is passed over the net : tineProtocol 6 (in) */
{
  UINT16    msgsize;
  UINT16    extsize;        /* extended string space */
  UINT16    mtu;            /* maximum transport unit for incoming data */
  UINT16    numblks;        /* total number of message blocks           */
  UINT16    blknum;         /* block number of this message             */
  UINT16    id;
  short     mode;
  short     blkid;          /* (incremented) block id */
  UINT32    pollingInterval;
  UINT32    starttime;
  CONTRACT contract;
} SubInfoPkt;

#define SUBINFO_SIZE (8*2+2*4+CONTRACT_SIZE)
#define SUB_MODE_LOCATION (PKTHDR_SIZE + 8)

typedef struct subscription5         /* struct is passed over the net : tineProtocol 5 (in) */
{
  UINT16    msgsize;
  UINT16    mtu;            /* maximum transport unit for incoming data */
  UINT8     numblks;        /* total number of message blocks           */
  UINT8     blknum;         /* block number of this message             */
  UINT16    id;
  short     mode;
  short     blkid;          /* (incremented) block id */
  UINT32    pollingInterval;
  UINT32    starttime;
  CONTRACT5 contract;
} SubInfoPkt5;

#define SUBINFO5_SIZE (6*2+2*4+CONTRACT5_SIZE)
#define SUB5_MODE_LOCATION (PKTHDR_SIZE + 8)
#define SUBMTU(np,tp) (((np) == IPX ? IPX_BUFFER_SIZE : UDP_BUFFER_SIZE) - \
                       ((tp) == 5 ? SUBINFO5_SIZE : SUBINFO_SIZE) - \
                       PKTHDR_SIZE)

extern ContractListStruct **ContractList;
extern ClnHdr **ClnTbl;         /* registered consumer list */

extern SOCKET  udpSrvSck;   /**< UDP Producer Socket */
extern SOCKET  udpClnSck;   /**< UDP Consumer Socket */
extern SOCKET  mcastClnSck; /**< UDP MCast/BCast Consumer Socket */
extern SOCKET  netSrvSck;   /**< UDP MCast Net Services Socket */
extern SOCKET  glbClnSck;   /**< UDP Global listening Socket */
extern SOCKET  udpSyncSck;  /**< UDP trivial link Socket */
extern struct sockaddr_in prod_name;

void clientCycle(void);

typedef struct
{
  char prpName[PROPERTY_NAME_SIZE];
  UINT32 prpSize;          /* maximum supported array size (output) */
  UINT32 prpSizeIn;        /* maximum supported array size (input)  */
  short prpFormat;         /* format (output) */
  short prpFormatIn;       /* format (input) */
  short prpAccessMode;     /* bit 0: read access, bit 1: write access */
  short prpArrayType;      /* output array type */
  char prpStructTag[TAG_NAME_SIZE]; /* output data structure tag */
  char prpDescription[64]; /* short description of property */
  int (*fcn)(CONTRACT *,char *,char *,double *);
  int runInSeparateThread; /* flag for potential slow pokes ... */
  int isFecProperty;       /* stock property refers to FEC and not the EQM */
} StockPropertyStruct;
#ifndef StockPropertyStructType
# define StockPropertyStructType StockPropertyStruct
#endif
extern int NrStockProperties;
extern StockPropertyStructType StockProperty[];
#define GetStockPropertyId(str) \
  getTableIndex(str,StockPropXRefTable,STOCKPROP_HASH_SIZE,(char *)StockProperty,NrStockProperties,OFFSETIN(StockPropertyStruct,prpName),sizeof(StockPropertyStruct))

typedef struct
{
  char modName[16];
  int  (*fcn)(CONTRACT *,char *,char *,double *);
  int  runInSeparateThread; /* flag for potential slow pokes ... */
  int  isEmbedded; /* is meta tag embedded in the property (else 'endswith') */
} StockModifierStruct;
#ifndef StockModifierStructType
# define StockModifierStructType StockModifierStruct
#endif
extern int NrStockModifiers;
extern StockModifierStructType StockModifier[];

typedef struct 
{
  CRS StkCrs;
  ContractListStruct *StkCls;
  int isThreadBusy;
  int isInfoLocked;
# ifdef MULTITHREADED
  int  (*fcn)(CONTRACT *,char *,char *,double *);
# ifdef UNIX
  pthread_mutex_t StkInfoMutex;
  pthread_mutex_t StkCallMutex[3];
# endif
  MXHANDLE hStkInfoMutex;
  MXHANDLE hStkCallMutexTriad[3];
  int hStkCallThread;
# else
  int hStkInfoMutex;
# endif
} StockRecourceSet;
extern StockRecourceSet gStockRecourceSet;

short IsMetaProperty(char *Property,char **baseProperty,char **metaProperty,int localOnly);
short IsStaticProperty(char *Property);

#ifndef MESSAGE_LOG_DEPTH
#define MESSAGE_LOG_DEPTH 500
#endif
typedef struct 
{
  time_t ts;
  int status;
  char msg[128];
} MsgType;
extern MsgType MsgTbl[MESSAGE_LOG_DEPTH];
extern int gMsgTblPtr;
int msglog(int status,char *text,...);
TINE_EXPORT int GetMsgTable(MsgType *msgtbl,int *nummsg);
TINE_EXPORT void ClearMsgTable(void);

#define MAX_ALIAS_ENTRIES 1024
#ifndef ALIAS_HASH_SIZE
# define ALIAS_HASH_SIZE 103
#endif
#define ALIAS_HASH_PAYOFF (ALIAS_HASH_SIZE/10)
typedef struct
{
  char name[DEVICE_NAME_SIZE];
  char alias[DEVICE_NAME_SIZE];
} ALIAS_TABLE;
extern ALIAS_TABLE *gAliasTable;
extern int nAliasTableEntries;
extern HashXRefStruct **AliasXRefTable;

int makeAliasXRefTable(void);
int GetAliasTable(char *eqm);
int getTableIndex(char *str,HashXRefStruct **xrf,int hshsiz,char *tbl,int tblsiz,int off,int reclen);
#define getAliasTableIndex(str) \
   getTableIndex(str,AliasXRefTable,ALIAS_HASH_SIZE,(char *)gAliasTable,nAliasTableEntries,OFFSETIN(ALIAS_TABLE,alias),sizeof(ALIAS_TABLE))
#define getAliasTableIndexFromName(str) \
   getTableIndex(str,NULL,0,(char *)gAliasTable,nAliasTableEntries,OFFSETIN(ALIAS_TABLE,name),sizeof(ALIAS_TABLE))

typedef void (*CYCBFCNP)(int,int,void *);
typedef struct cycTrgStruct
{
  CYCBFCNP fcn;
  char eqm[EQM_NAME_SIZE];
  char prpLst[256];
  void *ref;
  struct cycTrgStruct *nxt;
} CycTrg;
extern CycTrg *gCycTrgLst;
extern SINT32 gCycleNumber;
extern SINT32 gCycleOffset;
extern int useCycleTrigger;

/**
 * \fn TINE_EXPORT void SetSystemStampOffset(int cycleOffset)
 * \brief Establishes a system cycle offset
 * 
 * If a server's context has a registered 'CYCLER' then all read data will be tagged with the 
 * incoming system cycle number.  If it is known a priori that due to hard i/o latency the cycle
 * tag needs to be offset by some value, then this routine may be used to establish such an offset.
 *  
 * \param cycleOffset is the desired cycle offset (counts) to be applied to the incoming cycle number
 * from the registered CYCLER. (Default = 0).
 *
 * \see GetSystemStampOffset
 */
TINE_EXPORT void SetSystemStampOffset(int cycleOffset);

/**
 * \fn TINE_EXPORT int GetSystemStampOffset(void)
 * \brief Returns the registered system cycle offset
 * 
 * If a server's context has a registered 'CYCLER' then all read data will be tagged with the 
 * incoming system cycle number.  If it is known a priori that due to hard i/o latency the cycle
 * tag needs to be offset by some value, then this routine may be used to obtain the currently registered
 * cycle offset.
 *  
 * \return The currently registered system cycle offset (default = 0).
 *
 * \see SetSystemStampOffset
 */
TINE_EXPORT int GetSystemStampOffset(void);

/**
 * \fn TINE_EXPORT void SetSystemStampDelay(int cycleDelay)
 * \brief Establishes the system cycle delay
 * 
 * If a server's context has a registered 'CYCLER' then all read data will be tagged with the 
 * incoming system cycle number.  If it is known a priori that due to hard i/o latency the application of 
 * the cycle tag needs to be delayed by some value, then this routine may be used to establish such a
 * cycle delay value (in milliseconds).
 *  
 * \param cycleDelay is the desired cycle delay (milliseconds), which will must elapse before the incoming cycle number
 * from the registered CYCLER is to be applied to all readback data. (default = 0).
 *
 * \see GetSystemStampDelay
 */
TINE_EXPORT void SetSystemStampDelay(int cycleDelay);

/**
 * \fn TINE_EXPORT int GetSystemStampDelay(void)
 * \brief Returns the registered system cycle delay
 * 
 * If a server's context has a registered 'CYCLER' then all read data will be tagged with the 
 * incoming system cycle number.  If it is known a priori that due to hard i/o latency the application of 
 * the cycle tag needs to be delayed by some value, then this routine may be used to obtain the currently registered
 * cycle delay value (in milliseconds).
 *  
 * \return The currently registered system cycle delay (default = 0 msec).
 *
 * \see SetSystemStampDelay
 */
TINE_EXPORT int GetSystemStampDelay(void);

/* RegisterCycleTriggerFunction() */
/**
 * \fn TINE_EXPORT int RegisterCycleTriggerFunction(CYCBFCNP fcn,char *eqm,char *prpLst,void *reference)
 * \brief Registers a cycle trigger callback dispatch function
 * 
 * If a CYCLER is running in a server's context, then the server will receive 'Cycle Number' events
 * scheduled by the designated CYCLER server.
 * The cycle number will make use of the 'System Data Stamp' to tag all data sets obtained from the
 * server.  A server can also register a trigger function dispatch routine (or routines) to be called when a 'Cycle
 * Number' event occurs.  The dispatch routines will be called prior to setting the 'System Stamp' to the new
 * Cycle Number, which will be set following the execution of all dispatch routines.  Optionally, the server can
 * provide a property (or list of properties) to be scheduled following the dispatch execution.  This will ensure
 * that such properties will be called immediately following dispatch execution AND contain the most recent Cycle Number
 * as the 'System Data Stamp'.
 *  
 * \param fcn is a reference to the cycle trigger dispatch routine to be called following the reception of a new
 * Cycle Number.  This must have the prototype: void (*fcn)(int cycleNumber,int cycleStatus,void *reference).
 * Thus, the dispatch routine will be called with the current cycle number, the cycle status (possibly 'link_timeout' if
 * no cycle is received within the assigned globals heartbeat (see SetGlobalsHeartbeat)), and an optional 'reference',
 * which is a void pointer of the caller's choosing (and will be returned to the caller in the dispatch routine).
 * \param eqm the local equipment module name of the desired central server (e.g. "BPMEQM")
 * \param prpLst is the property or properties which are to be 'scheduled' following the execution of the dispatch
 * routine.  If more than a single property is to be scheduled, this should be a string containing a comma separated
 * list.
 * \param reference is a caller supplied void pointer which will be returned to the called in the dispatch
 * routine.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \b Example:
 *
 * \include eg_cycletrigger.c
 *
 * \see UnregisterCycleTriggerFunction
 */
TINE_EXPORT int RegisterCycleTriggerFunction(CYCBFCNP fcn,char *eqm,char *prpLst,void *reference);

/* FindServerOnNetwork() */
/**
 * \fn TINE_EXPORT int UnregisterCycleTriggerFunction(CYCBFCNP fcn,void *reference)
 * \brief Unregisters a previously registered cycle trigger callback dispatch function
 * 
 * If a cycle trigger event dispatch routine is no longer required, it can be unregistered using this 
 * routine.
 *  
 * \param fcn is a reference to the previously registered cycle trigger dispatch routine which is to
 * be removed.
 * \param reference is a caller supplied void pointer which will be returned to the called in the dispatch
 * routine.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see RegisterCycleTriggerFunction
 */
TINE_EXPORT int UnregisterCycleTriggerFunction(CYCBFCNP fcn,void *reference);

#define CDI_LNK_TBL 0x1000000
#define MAKE_CDI_LINKID(id) (id < 0 ? id : ((id) | CDI_LNK_TBL))
#define GET_CDI_LINKID(id) ((id) & CDI_LNK_TBL ? (id) & ~(CDI_LNK_TBL) : -1)
extern TINE_EXPORT void (*cdiSystemInitHook)(void);
extern TINE_EXPORT int (*cdiCloseLinkHook)(int);
extern TINE_EXPORT char *(*getCdiLinkNameHook)(char *,char *);
extern TINE_EXPORT int (*cdiExecLinkHook)(char *,char *,DTYPE *,DTYPE *,short,UINT16);
extern TINE_EXPORT int (*cdiAttachLinkHook)(char *,char *,DTYPE *,DTYPE *,short,int,TCBFCNP,int,UINT32);
extern TINE_EXPORT void (*cdiSystemFreeHook)(void);

#ifdef SERVER_POLLING_MODE
#define ServerCycleMode    POLLING
#else
#define ServerCycleMode    SELECTING
#endif
extern int RequireAcknowledgments;
extern int AveBusyTime;
extern struct timeval thisCycleTime;
extern BYTE *BrdcstBuffer;
extern UINT32 TCPcumcnt,TCPinitcnt;
extern time_t gSystemStartupTime;
extern int hConn, hControl;
extern WORD ISwap(WORD w);
extern int ipLoaded,ipxLoaded,numErr;
extern char gUserName[20];      /* login name of PC user */
extern TINE_EXPORT char gFecName[FEC_NAME_SIZE+4];
extern BYTE *clientbuf;
extern BYTE *WorkArea;
extern BYTE *tmpWorkArea;
extern UINT32 WorkAreaSize;
extern UINT32 tmpWorkAreaSize;

extern char gFecAddrBuffer[30],gClientAddrBuffer[30],fecprot[16];
extern int gListenForMCasts;
extern int foregroundTTY;
extern UINT32 CMDcount;

extern ContractListStruct *ContractListItem;
extern ConTblEntry **conTbl;
extern GrpTblEntry *grpTbl;
extern int ReturnToClientCycle;
extern int ReturnToServerCycle;
extern int nConnectionTableEntries;
extern struct sockaddr_in srv_name,brdcst_name;
extern SOCKET tcpSrvSck,tcpStrmSck,dbgSrvSck;
extern struct timeval TCPto, fec_to;
extern int CheckDeniedList;
extern BYTE spxSrvSckB[2];
extern const BYTE ipxClnSckB[2];
extern BYTE ipxSrvSckB[2];
extern const BYTE ipxMcastSckB[2];
extern char PATH[];
extern FecDataStruct *FecTbl, *NameServer;
extern ExpDataStruct *SrvTbl;
extern FecInfoStruct gFecInfo;
extern int numNameServers,curNameServer;
extern int numSrvTblEntries,numFecTblEntries,FecTblSize;
extern UINT32 TCPcount,UDPcount,SPXcount,IPXcount,BRDcount;
extern UINT32 nConnectionTimeouts,nArrivals;
extern UINT32 nPartialTransfers;
extern short gNumEqmBkgTasks;
extern time_t *tnull;
extern TINE_EXPORT char gDeviceContext[CONTEXT_NAME_SIZE+4];
extern TINE_EXPORT char gDeviceSubSystem[SUBSYSTEM_NAME_SIZE+4];
extern UINT32 gContractMisses;
extern UINT32 gContractDelays;
extern UINT32 gClientMisses;
extern UINT32 gClientReconnects;
extern UINT32 gClientRetries;
extern UINT32 gBurstLimitReachedCount;
extern UINT32 gSingleLinkCount;
extern int gCycleCounts, gMaxCycleCounts;
extern int gLastMinuteInitializationDone;
extern int max_clnrcvbuf;
extern int max_srvrcvbuf;
extern int max_clnsndbuf;
extern int max_srvsndbuf;
extern TINE_EXPORT int gRetardSingleContractRemoval;
extern TINE_EXPORT int ServerExitCondition;
extern TINE_EXPORT int ServerInitialized;
extern TINE_EXPORT int ServerWaiting;
extern int acquireDeviceNamesFromProperties;
extern int canSendGlobals;
extern char *ltzname[2];
extern char tExecLinkReturnedCompletion[STATUS_SIZE+4];
extern int tExecLinkReturnedDataSize;
extern int gUsingLegacyAddrCalls;
extern int gTineServicesProtocol;
extern int gEqmFcnContractTableId;
extern int gEnforceMcaAccess;
extern int gClientCycleIdle;
extern int gServerCycleIdle;

TINE_EXPORT void SetServerExitCondition(int value);
/**
 \internal if reattaching to multicast group is relevant
 */
void reAttachToServiceGroup(void);
int SystemStartCycleTrigger(void);
int makeValidPollingInterval(int pollingRate, int baseMode);
char *makePropertyFileName(char *eqm,int usesubdir,char *prp,char *decoration);
void sendPrpSignal(ExportPropertyListStruct *prp,int signal,int conId,int sts);
void prepOutgoingData(BYTE *dst,BYTE *src,UINT32 dsiz,short fmt);
void prepIncomingData(BYTE *dst,BYTE *src,UINT32 dsiz,UINT32 rsiz,short fmt,ConTblEntry *c);
int prepSubHeader(int tineProtocol,SubInfoPkt *pSub,BYTE *pSubstream,int *swallowError);
int prepOutgoingSubscription(int tineProtocol,ConTblEntry *c,char *eqm,int datasize,int subid,BYTE *pSubstream);
int prepIncomingSubscription(int tineProtocol,PrdrHdr *pPHdr,BYTE *pHdrstream);
int prepProducerHeader(int tineProtocol,int xferReason,ClnLst *cn,CONTRACT *ct,int returnCode,int dstamp,int sstamp,
                       UINT32 tv_sec,UINT32 tv_msec,UINT16 msgsize,BYTE *pPHdrstream);
#define SUB_STEPSIZE(p) ((p) == 5 ? SUBINFO5_SIZE : SUBINFO_SIZE)
#define PHDR_STEPSIZE(p) ((p) == 5 ? PHDR5_SIZE : PHDR_SIZE)

NAME64 *GetEqpDeviceNames(char *eqm,char *prpname,int *listSize,EqmDeviceRdrListType **rdrnames,EqmDevicePropertyList **prplist,NAME64 **dsclist);
int IsIllegalProperty(char *eqmName,char *prpName,char *dev,ClnHdr *cln,short access,
                      ExportListStruct **expLstElem,ExportPropertyListStruct **prpLstElem,char *stsString);
int doContract(ContractListStruct *cl);
void doGCast(struct timeval *thisTime);
int GetPrpId(char *eqm,char *prpName);
void devNamesInitialization(void);
void ExitEQPModules(void);
void checkTasklist(void);
int GetDeviceListFromProperty(char *eqm,char *prp,NAME64 *devlst,UINT32 *devlen);
char *getFirstDevice(char *eqm,char *prpname);
ExportPropertyListStruct *AssignDeviceListFromProperty(char *eqm,char *prp);
int prepSubRequest(int isENS);
int sendSubRequest(int fecIdx,SOCKET sck,int inetProtocol,BYTE *data);
unsigned short execRequest(ClnHdr *cln);
#define SCHED_INTERN 0
#define SCHED_EXTERN 1
int doScheduler(struct timeval *thisTime,int timeout,int source,ClnHdr *tgtcln,ContractListStruct *tgtcon);
int doDelivery(struct timeval *thisTIme,ClnHdr *tgtcln,ContractListStruct *tgtcon);
int sendToPeer(ClnHdr *c,BYTE *payload,int msgtype);
int doRequest(int num, int *i);
int getConsumerRequest(ClnHdr *cln,BYTE *buf,int packetsize);
int getConTblEntry(DTYPE *dout,DTYPE *din,char *devcontxt,char *devgrp,char *devname,char *property,short access,int reuseIndex);
void displayData(int format, int fec, int dir);
int dumpConsumerList(void);
int dumpContractList(void);
int dumpMessageList(void);
int dumpEqmList(void);
int dumpEqmProperties(char *eqm);
int dumpEqmDevices(char *eqm);
int dumpConnectionList(void);
void dumpRegisteredUsers(void);
int dumpConnectionGroups(void);
int dumpSemaphores(void);
void prepDataDump(char *buf,char *data,int siz,short fmt);
void consumeData(UINT32 bufsize,BYTE *buf);
int addFecToList(char   *name);
int initFecTable(void);
int datacmp(BYTE *d1,BYTE *d2,int siz,int fmt,double t);
void getGlobalUpdate(UINT32 addr,BYTE *buf);
void quitServer(void);
void hardDelay(int msec);
int isMemberLockSet(ExportListStruct *el,ClnHdr *cln);
void lockToExclusiveRead(ClnHdr *cln);
int removeClientFromContract(ClnLst *c,int tid);
int removeClientTableEntry(int tid);
int removeContractTableEntry(int tid);
int isExclusiveProperty(ExportListStruct *el,char *prp);
int isAllowedUser(char *eqm,char *usr);
int isDeniedUser(char *usr);
int isAllowedPropertyAccess(const char *eqm,const char *prp,ClnHdr *cln);
int isAllowedDeviceAccess(const char *eqm,const char *dev,ClnHdr *cln);
ClnHdr *getNetworkClientFromList(void);
ClnHdr *LocateConsumerInList(struct sockaddr_in *ip,IPXAddress *nwAddr,BYTE *immAddr);
TINE_EXPORT UINT32 ElfHash(const unsigned char *key);
void makeEqpDevicesHashTable(ExportListStruct *el);
int makeStockPropXRefTable(void);
#define GATE_BOOLEAN 0x01
#define GATE_FILTER 0x02
short IsGatedModifier(CONTRACT *con,char *baseProperty,UINT32 *gate);
short IsStockProperty(char *Property);
short IsFecProperty(char *Property);
int hasVolatileDataSize(int fmt,char *tag);
int lookupPropertyName(ExportListStruct *el,char *prpName,char *devName,ExportPropertyListStruct **prp);
char *lookupPropertyId(ExportListStruct *el,short id);
int sndRegisteredExports(struct timeval *tv);
void ParsePropertyDescription(char *str,ExportPropertyListStruct *prp);
void ParseRedirector(char *str,PrpRedirBlk *prd);
ExportListStruct *getExportListItem(char *eqm);
ExportListStruct *getExportListItemFromId(int id);
int SystemStartGlobalSynchronization(void);
TINE_EXPORT int reformat(BYTE *dstData,int dstFmt,BYTE *srcData,int srcFmt,int num);
int getSrvAddress(DTYPE *sa,char *eqm,char *devsrv);
TINE_EXPORT int GetServerInitialization(void);

int IsInForeground(void);
void checkIPXNets(struct timeval *tv);
int GetSPXStream(WORD ID,ClnHdr *cln);
int initIPX(void);
void CleanupIPX(struct timeval *tv);
void closeIPXServer(void);
void AcceptIPX(struct timeval *tv);
void AcceptSPXStream(void);
void CloseSPXStream(WORD ID);
int sendIPXData(ClnHdr *cln, BYTE *data);
int sendIPXgcast(BYTE *buf,int size);
void fulfillSPXStreamRequest(WORD ID);
ClnHdr *getConsumerIPXAddress(void);
char *ipxnet_ntoa(BYTE *node);
void dumpIPXNets(void);

int _recvGlobal(const char *keyword,DTYPE *dout,void *callback,UINT32 callbackID,void *reference);
int _attachLink(const char *devNam,const char *devProp,DTYPE *dout,DTYPE *din,short access,
	       int pollingRate,void *callback,int mode,UINT32 callbackID,void *reference,int reuseIndex);
void _closeLink(int linkId);
void NotifyClient(int i,int cc);
void NotifyGlobalClient(int i,int cc);
int doTCPClient(int fec);
int setupUdpSocket(SOCKET *s,unsigned short port,int initMultiCast,int rcvbuf,int sndbuf);
int sendUDPBuffer(struct sockaddr_in *client, int msgsize);
int connectTCP(ConTblEntry *c,int port);
int initIP(void);
int initClient(void);
int sendIPData(ClnHdr *cln,BYTE *data,int inetProtocol);
int sendStreamData(ClnHdr *cln,BYTE *data);
int sendIPgcast(BYTE *buf,int size);
void fulfillTCPStreamRequest(int socket);
ClnHdr *getConsumerIPAddress(SOCKET sck,int inetProtocol);
int redirectLink(char *con,char *srv,char *prp,char *dev);
void AcceptTCPStream(void);
void dumpIPNets(void);
void AcceptIP(struct timeval *tv);
void CleanupIP(struct timeval *tv);
void closeIPServer(void);
int IsMemberIPNets(int num,struct sockaddr_in *lst,struct sockaddr_in *ipaddr);
int tUDPCall(int fecindex,BYTE *buf,int timout);
int tIPXCall(int fecindex,BYTE *buf,int timout);
int tExecLink(char *fec,char *eqm,char *dev,char *prp,DTYPE *dout,DTYPE *din,short access,short timout);
void checkClientNotification(void);
void checkSubscriptions(int isENS);
int readSelfTest(char *eqm,char *textbuf,int sizeInBytes);
int readFecLog(char *textbuf,int sizeInBytes);
int readFecLogFile(char *fname,char *textbuf,int sizeInBytes);
int writeFecLogFile(char *fname,char *textbuf,int sizeInBytes);
int readFecFile(char *fname,char *buf,int *sizeInBytes);
int writeFecFile(char *fname,char *buf,int sizeInBytes);
int scanFecLogFiles(char *path,char *subpath,NAME64 *namebuf,int *numfiles,int numoffset);
void applyFecLogPath(char *path,int pathlen);
void checkAccessLocks(time_t t);
int isDeniedUser(char *usr);
TINE_EXPORT void haltServer(void);
void freeServerMemory(void);
TINE_EXPORT void FreeEQPModules(void);
TINE_EXPORT int dbglog(char *text,...);
int _datcmp(char *d1,char *d2,int dsz,short fmt,DataToleranceFactor *dtf);
int isSpectrumArray(char *eqm,char *prpName);
int isChannelArray(char *eqm,char *prpName);
void getWildcardTokens(char *wildcardString,char *beginToken,char *endToken,char *midToken);
int wildcardMatches(char *devName,char *beginToken,int beginLen,char *endToken,int endLen,char *midToken,int midLen);
int isContiguous(char *eqm,char *devName,NAME64 *devList,int *start,int *len);
int isLocal(char *eqm,char *devName,char *prpName);
ExportPropertyListStruct *GetPropertyListStruct(char *eqm,char *prpName,char *devName);
int queryBitfieldFormat(char *srv,char *tag,NAME16I *bffq,int *bffqsiz);
int callEqmInSeparateThread(char *eqm,char *prp,int access);
extern int gClnRecvQueueDepth;
int callEqm(ContractListStruct *cl,char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
int IsPropertyCalledInSeparateThread(char *eqm,char *property);
void startFailoverMonitor(char *eqm);
void _cbXRef(int id, int cc);
void _cbMcaXRef(int id, int cc,void *ref);
int getCbXRefTblId(void);
void flushContractTable(void);
void sendContractSignal(const char *eqm,const char *dev,const char *prp,int signal,int sts);

ReLnkTblEntry *getReLnkLstItem(char *ctx,char *srv,char *dev,char *prp,int siz);
void appendReLnkLst(ConTblEntry *c,char *tag,int fmt);
int relinkLink(char *ctx,char *srv,char *prp,char *dev,char **pkey,DTYPE *dout);
int reAttachLink(int i,int isStructField);
void flushReLnkLst(void);

int isBitfieldAccess(DTYPE *d);
int isBitfieldProperty(char *srv,char *tag,char *prp,char *fld);
UINT32 getBitfieldValue(int fmt,BYTE *dptr);
void mapBitfieldToTarget(ConTblEntry *c);
int splitFullName(const char *fullName,char *srvName,char *devName);
int GetIntValue(DTYPE *d,SINT32 *ival);

int isTaggedStructAccess(DTYPE *d);
int isStructProperty(char *tag,char *prp,char *fld);
void mapStructToTarget(ConTblEntry *c);

TINE_EXPORT int tokenizeFullName(char *fullName,char **ctx,char **srv,char **dev);
TINE_EXPORT void ResetLastMinuteInitialization(void);
TINE_EXPORT void resetLastMinuteInitialization(void);

/* GetCallbackGroup() */
/**
 * \fn TINE_EXPORT GrpTblEntry *GetCallbackGroup(size_t id)
 * \brief Returns a reference to the callback Group Table Entry associated with the
 * identifier supplied
 * 
 * \param id is the group callback id (effectively the reference pointer to the callback
 * function associated with the group.
 *  
 * \return a reference to the group table entry (GrpTblEntry *) if the given id corresponds
 * to a registered callback group, or NULL if not found.
 *
 * \see GetGroupMemberList(), GrpTblEntry, GrpMember
 */
TINE_EXPORT GrpTblEntry *GetCallbackGroup(size_t id);

/* GetGroupMemberList() */
/**
 * \fn TINE_EXPORT GrpMember *GetGroupMemberList(GrpTblEntry *grp)
 * \brief Returns a linked list of the group members for the callback group supplied
 * 
 * \param grp is a reference to the group table entry (GrpTblEntry *) whose member list
 * is desired.  
 *  
 * \return a reference to a linked list of the group members for the callback group supplied
 *
 * \see GetCallbackGroup(), GrpTblEntry, GrpMember
 */
TINE_EXPORT GrpMember *GetGroupMemberList(GrpTblEntry *grp);

/* FindServerOnNetwork() */
/**
 * \fn TINE_EXPORT int FindServerOnNetwork(char *context,char *eqmName,char *exportName,FecDataStruct *fec,ExpDataStruct *srv)
 * \brief Issues a multicast (or broadcast) to which configured TINE central servers respond
 * 
 * Diskless (or otherwise 'in-a-box') servers which need to find central servers such as the
 * Equipment Name Server, Central Alarm Server, or Post-Mortem Server, etc. can issue this
 * call to ascertain the address.
 *  
 * \param context (optional) the Context of the desired central server
 * \param eqmName the local equipment module name of the desired central server (e.g. "ENSEQM")
 * \param exportName the exported device server name of the desired central server. Note either
 * exportName or eqmName, but not both are optional input parameters.
 * \param fec (optional) is the returned FEC address structure of the server responding to the
 * call
 * \param srv (optional) is the returned Device Server Data address structure of the server responding
 * to the call.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
TINE_EXPORT int FindServerOnNetwork(char *context,char *eqmName,char *exportName,FecDataStruct *fec,ExpDataStruct *srv);

/**
 * \fn TINE_EXPORT int GetCallerInformation(char *eqm,ClnInfo *clnInfoList,int *num)
 * \brief Returns the user name, network address and other information of all callers interested in the current contract
 * 
 * \param eqm (in) is the equipment function name (local name) for which the exported name
 * is desired.
 * \param clnInfoList (out) if non-NULL is a reference to an array of ClnInfo objects to be filled with the caller list
 * \param num (in/out) is a reference to an int value which initially contains the size of the ClnInfoList array
 * used in the parameter list. Upon return of the call, it contains the actual number of callers.
 *
 * \note The clnInfoList supplied must point to buffer with enough room to hold at least the input 'num'
 * entries, otherwise a system crash is likely.
 *
 * \return 0 if successful, or a TINE completion code
 *
 * \see GetCaller()
 *
 * \b Example:
 *
 * \include eg_GetCallerInformation.c
 */
TINE_EXPORT int GetCallerInformation(char *eqm,ClnInfo *clnInfoList,int *num);

TINE_EXPORT void SetRunAsServer(void);
extern int gIsRunningAsServer;

TINE_EXPORT void SetServerCycleIdle(int value);
TINE_EXPORT int GetServerCycleIdle(void);
TINE_EXPORT void SetClientCycleIdle(int value);
TINE_EXPORT int GetClientCycleIdle(void);

extern int gSubRenewalLength;
extern int gSubRenewalMultiplier;

TINE_EXPORT void SetRespondToServiceRequests(int value);
TINE_EXPORT int GetRespondToServiceRequests(void);

/**
 * \fn TINE_EXPORT void SetSystemSubscriptionRenewalLength(int value)
 * \brief Sets the contract subscription renewal length
 * 
 * Persistent contracts established by a client calling one of the AttachLink()
 * family 'subscribe' for the contract on the server.  A subscription
 * is only valid for a number of responses and must be renewed by the client.
 * This of course happens automatically in the TINE kernel and is a guarantee 
 * that clients which 'disappear' ungracefully do not persist indefinitely.
 * The 'number of responses' is known as the subscription renewal length.
 * This value is adjusted according to a client's desired polling interval
 * (i.e. the default value at 1 Hz, 60, is augmented to higher values when
 * polling for instance at 10 Hz).
 * The default value might under some circumstance be deemed to be too 
 * small. e.g. a server knows that properties are being 'scheduled'
 * (see _SystemScheduleProperty()) at many Hz, which would cause the 
 * subscription counter to descrease more rapidly than the client's 
 * polling interval would otherwise dictate.  In such cases the subscription
 * 'renewals' would be far more frequently than an efficient data
 * transmission would warrant.  The default value for subscription
 * renewals can be adjusted with this call.
 *
 * \param value is the desired contract renewal length.
 * (mininum = default = 60; maximum = 32767)
 *
 * \see GetSystemSubscriptionRenewalLength()
 *
 */
TINE_EXPORT void SetSystemSubscriptionRenewalLength(int value);

/**
 * \fn TINE_EXPORT int GetSystemSubscriptionRenewalLength(void)
 * \brief Gets the current contract subscription renewal length
 * 
 * Persistent contracts established by a client calling one of the AttachLink()
 * family 'subscribe' for the contract on the server.  A subscription
 * is only valid for a number of responses and must be renewed by the client.
 * This of course happens automatically in the TINE kernel and is a guarantee 
 * that clients which 'disappear' ungracefully do not persist indefinitely.
 * The 'number of responses' is known as the subscription renewal length.
 * This value is adjusted according to a client's desired polling interval
 * (i.e. the default value at 1 Hz, 60, is augmented to higher values when
 * polling for instance at 10 Hz).
 * The default value might under some circumstance be deemed to be too 
 * small. e.g. a server knows that properties are being 'scheduled'
 * (see _SystemScheduleProperty()) at many Hz, which would cause the 
 * subscription counter to descrease more rapidly than the client's 
 * polling interval would otherwise dictate.  In such cases the subscription
 * 'renewals' would be far more frequently than an efficient data
 * transmission would warrant.  The default value for subscription
 * renewals can be retrieved with this call.
 *
 * \return the current setting of the contract renewal length
 *
 * \see SetSystemSubscriptionRenewalLength()
 *
 */
TINE_EXPORT int GetSystemSubscriptionRenewalLength(void);

/**
 * \fn TINE_EXPORT void SetPreSystemInitFunction(SYSTSKP fcn)
 * \brief Sets a user-specific initialization routine to be executed prior to 
 * server initialization.  
 *
 * \param fcn is the function to be called prior to system initialization.
 * (prototype void fcn(void)). Typically one places parameter settings which 
 * differ from the default values in such a routine, so that they have 
 * an effect on the server initialization.  Likewise, calls such as 
 * RegisterFecInformation() should be made in such a routine.
 *
 * \see SetPostSystemInitFunction()
 *
 */
TINE_EXPORT void SetPreSystemInitFunction(SYSTSKP fcn);

/**
 * \fn TINE_EXPORT void SetPostSystemInitFunction(SYSTSKP fcn)
 * \brief Sets a user-specific initialization routine to be executed following 
 * server initialization.  
 *
 * \param fcn is the function to be called following system initialization.
 * (prototype void fcn(void)). Typically one places registration calls 
 * such as RegisterEquipmentModule(), etc. in such a routine. 
 *
 * \see SetPreSystemInitFunction()
 *
 */
TINE_EXPORT void SetPostSystemInitFunction(SYSTSKP fcn);

#define getMultiplierFromRate(r) ( (r) > 0 && (r) < 1000 ? 1000/(r) : 1)
#define getThisSubscriptionRenewalLength(r) (\
  gSubRenewalLength != CTR_RENEWAL ? gSubRenewalLength : CTR_RENEWAL * getMultiplierFromRate(r)\
)
#define getThisSubscriptionRenewalMultiplier(m,r) (\
  (m) > 1 ? (m) : getMultiplierFromRate(r)\
)

extern int gSystemDataStamp;
extern int gSystemDataStampRb;
extern int gSystemDataStampDly;
extern struct timeval gSystemDataStampTs;
void applySystemDataStamp(void);

TINE_EXPORT void SetEnforceMultiChannelAccess(int value);
TINE_EXPORT int GetEnforceMultiChannelAccess(void);

TINE_EXPORT FecDataStruct *GetFecDataFromCache(char *context,char *expName);
TINE_EXPORT int GetFECAddr(char *fecName,FecDataStruct *fecaddr);
TINE_EXPORT int GetServerAddr(char *context,char *expName,ExpDataStruct *srvaddr,FecDataStruct *fecaddr);
TINE_EXPORT int GetConnectionAddr(char *context,char *expName,char *devName,char *prpName);
TINE_EXPORT void CloseClientsEx(int closesockets,int quickExit);
TINE_EXPORT void CloseClients(int closesockets);
#define closeClients(a1) CloseClients(a1)
TINE_EXPORT void CloseServers(int closesockets);
#define closeServers(a1) CloseServers(a1)
TINE_EXPORT void UseMultiThreadedCycler(int value);
TINE_EXPORT void UseSynchronousNotification(int value);
int getClockTickMilliseconds(void);
extern int heapCorrupted;
#ifdef WIN32
# define assertHeapOkay(t,d,p) {\
  if (_heapchk() != _HEAPOK && !heapCorrupted) {\
    feclog("%s %s %s : memory overwrite has occurred",t,d,p);\
    msglog(memory_overwrite,"%s %s %s: memory overwrite has occurred",t,d,p);\
    heapCorrupted = TRUE;\
  }\
}
#else
# define assertHeapOkay(t,d,p) ;
#endif
#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif  /* SRVCORE_H_ */
